# -*- coding: utf-8 -*-

import time
import logging
import threading
from threading import Thread

from litedfs.data.utils.task_cache import TaskCache
from litedfs.data.utils.registrant import Registrant
from litedfs.data.utils.common import delete_file, delete_block
from litedfs.data.config import CONFIG

LOG = logging.getLogger(__name__)


class StoppableThread(Thread):
    """
    Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition.
    """

    def __init__(self):
        super(StoppableThread, self).__init__()
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.isSet()


class TaskProcesser(StoppableThread):
    def __init__(self, pid):
        StoppableThread.__init__(self)
        Thread.__init__(self)
        self.pid = pid

    def run(self):
        LOG = logging.getLogger("task_processer")
        LOG.info("TaskProcesser(%03d) start", self.pid)
        try:
            while True:
                if not self.stopped():
                    try:
                        if not TaskCache.empty():
                            task = TaskCache.pop()
                            if task is not None:
                                if task["command"] == "delete":
                                    if "block" not in task:
                                        delete_file(task["name"])
                                    else:
                                        delete_block(task["name"], task["block"])
                                elif task["command"] == "replicate":
                                    Registrant.instance().replicate_block(task["name"], task["block"], task["ids"])
                            else:
                                time.sleep(0.5)
                            LOG.info("TaskProcesser(%03d) process task: %s", self.pid, task)
                        else:
                            time.sleep(5)
                    except Exception as e:
                        LOG.exception(e)
                else:
                    LOG.info("TaskProcesser(%03d) exit by siginal", self.pid)
                    break
        except Exception as e:
            LOG.exception(e)
        LOG.info("TaskProcesser(%03d) exit", self.pid)



# -*- coding: utf-8 -*-

import os
import sys
import json
import logging
from io import BytesIO

cwd = os.path.split(os.path.realpath(__file__))[0]
sys.path.insert(0, os.path.split(cwd)[0])

from litedfs.tool.litedfs import strings_md5sum, file_md5sum, bytes_io_md5sum
from litedfs.name import logger

LOG = logging.getLogger(__name__)


if __name__ == "__main__":
    logger.config_logging(file_name = "test_md5_methods.log",
                          log_level = "DEBUG",
                          dir_name = os.path.join(cwd, "logs"),
                          day_rotate = False,
                          when = "D",
                          interval = 1,
                          max_size = 20,
                          backup_count = 5,
                          console = True)

    LOG.debug("test start")
    
    try:
        LOG.debug("%s", strings_md5sum(["1", "2", "3"]))
        LOG.debug("%s", file_md5sum("./test_md5_methods.py"))

        fp = open("./test_md5_methods.py", "rb")
        content = BytesIO()
        content.write(fp.read())
        content.seek(0)
        LOG.debug("%s", bytes_io_md5sum(content))

    except Exception as e:
        LOG.exception(e)

    LOG.debug("test end")



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

import os
import re
import sys
import json
import time
import hashlib
import argparse
import logging
import random
from io import BytesIO

import requests

from litedfs.version import __version__
from litedfs_client.client import LiteDFSClient


BUF_SIZE = 65536


def print_table_result(data, fields, args):
    fields.insert(0, "#")
    field_length_map = {}
    lines = []
    num = 1
    column_max_width = args.column_width
    for field in fields:
        field_length_map[field] = len(field)
    for item in data:
        line = []
        for field in fields:
            if field == "#":
                line.append(str(num))
            else:
                v = str(item[field]) if field in item else ""
                v_len = len(v)
                if column_max_width > 0 and v_len > column_max_width:
                    v_len = column_max_width
                    v = v[:v_len]
                if v_len > field_length_map[field]:
                    field_length_map[field] = v_len
                line.append(v)
        lines.append(tuple(line))
        num += 1
    field_length_map["#"] = len(str(num))
    format_str = ""
    for field in fields:
        field_len = field_length_map[field]
        if field == "#":
            format_str += "%" + " %s" % field_len + "s | "
        else:
            format_str += "%" + "-%s" % field_len + "s | "
    format_str = format_str[:-3]
    print(format_str % tuple(fields))
    for line in lines:
        print(format_str % line)


def file_md5sum(file_path):
    md5 = hashlib.md5()
    with open(file_path, 'rb') as fp:
        while True:
            data = fp.read(BUF_SIZE)
            if not data:
                break
            md5.update(data)
    return md5.hexdigest()


def bytes_io_md5sum(fp):
    md5 = hashlib.md5()
    while True:
        data = fp.read(BUF_SIZE)
        if not data:
            break
        md5.update(data)
    return md5.hexdigest()


def bytes_md5sum(b):
    md5 = hashlib.md5()
    md5.update(b)
    return md5.hexdigest()


def strings_md5sum(l):
    md5 = hashlib.md5()
    for s in l:
        if s:
            md5.update(s.encode("utf-8"))
    return md5.hexdigest()


def main():
    parser = argparse.ArgumentParser(prog = 'ldfs')

    # common arguments
    parser.add_argument("address", help = "name node address, host:port")
    parser.add_argument("-w", "--column_width", help = "column max width", type = int, default = 0)
    parser.add_argument("-v", "--version", action = 'version', version = '%(prog)s ' + __version__)
    subparsers = parser.add_subparsers(dest = "object", help = 'sub-command help')

    # operate with file
    parser_file = subparsers.add_parser("file", help = "operate with file API")
    subparsers_file = parser_file.add_subparsers(dest = "operation", help = 'sub-command file help')

    parser_file_create = subparsers_file.add_parser("create", help = "create file")
    parser_file_create.add_argument("-l", "--local-path", required = True, help = "local file path", default = "")
    parser_file_create.add_argument("-r", "--remote-path", required = True, help = "remote file path", default = "")
    parser_file_create.add_argument("-R", "--replica", help = "replica count", type = int, default = 1)

    parser_file_delete = subparsers_file.add_parser("delete", help = "delete file")
    parser_file_delete.add_argument("-r", "--remote-path", required = True, help = "remote file path", default = "")

    parser_file_move = subparsers_file.add_parser("move", help = "move file")
    parser_file_move.add_argument("-s", "--source-path", required = True, help = "source file path", default = "")
    parser_file_move.add_argument("-t", "--target-path", required = True, help = "target directory path", default = "")

    parser_file_rename = subparsers_file.add_parser("rename", help = "rename file")
    parser_file_rename.add_argument("-r", "--remote-path", required = True, help = "remote file path", default = "")
    parser_file_rename.add_argument("-n", "--new-name", required = True, help = "new file name", default = "")

    parser_file_update = subparsers_file.add_parser("update", help = "update file")
    parser_file_update.add_argument("-r", "--remote-path", required = True, help = "remote file path", default = "")
    parser_file_update.add_argument("-R", "--replica", required = True, help = "replica count", type = int, default = 1)

    parser_file_download = subparsers_file.add_parser("download", help = "download file")
    parser_file_download.add_argument("-l", "--local-path", required = True, help = "local file path", default = "")
    parser_file_download.add_argument("-r", "--remote-path", required = True, help = "remote file path", default = "")

    parser_file_info = subparsers_file.add_parser("info", help = "get file's info")
    parser_file_info.add_argument("-r", "--remote-path", required = True, help = "remote file path", default = "")

    # operate with directory
    parser_directory = subparsers.add_parser("directory", help = "operate with directory API")
    subparsers_directory = parser_directory.add_subparsers(dest = "operation", help = 'sub-command file help')

    parser_directory_create = subparsers_directory.add_parser("create", help = "create directory")
    parser_directory_create.add_argument("-r", "--remote-path", required = True, help = "remote directory path", default = "")

    parser_directory_delete = subparsers_directory.add_parser("delete", help = "delete directory")
    parser_directory_delete.add_argument("-r", "--remote-path", required = True, help = "remote directory path", default = "")

    parser_directory_move = subparsers_directory.add_parser("move", help = "move directory")
    parser_directory_move.add_argument("-s", "--source-path", required = True, help = "source directory path", default = "")
    parser_directory_move.add_argument("-t", "--target-path", required = True, help = "target directory path", default = "")

    parser_directory_rename = subparsers_directory.add_parser("rename", help = "rename directory")
    parser_directory_rename.add_argument("-r", "--remote-path", required = True, help = "remote directory path", default = "")
    parser_directory_rename.add_argument("-n", "--new-name", required = True, help = "new directory name", default = "")

    parser_directory_list = subparsers_directory.add_parser("list", help = "list directory's children")
    parser_directory_list.add_argument("-r", "--remote-path", required = True, help = "remote directory path", default = "")
    parser_directory_list.add_argument("-o", "--offset", help = "list offset", type = int, default = 0)
    parser_directory_list.add_argument("-l", "--limit", help = "list limit", type = int, default = 0)
    parser_directory_list.add_argument("-f", "--exclude-file", help = "exclude file", action = "store_false")
    parser_directory_list.add_argument("-d", "--exclude-directory", help = "exclude directory", action = "store_false")

    # operate with cluster
    parser_cluster = subparsers.add_parser("cluster", help = "operate with cluster API")
    subparsers_cluster = parser_cluster.add_subparsers(dest = "operation", help = 'sub-command cluster help')

    parser_cluster_info = subparsers_cluster.add_parser("info", help = "cluster's info")
    parser_cluster_info.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")

    # operate with path(file or directory)
    parser_path = subparsers.add_parser("path", help = "operate with path API")
    subparsers_path = parser_path.add_subparsers(dest = "operation", help = 'sub-command path help')

    parser_path_info = subparsers_path.add_parser("info", help = "get path's info")
    parser_path_info.add_argument("-r", "--remote-path", required = True, help = "remote file/directory path", default = "")

    args = parser.parse_args()

    try:
        address = args.address
        object = args.object
        operation = args.operation
        url = "http://%s/%s/%s" % (address, object, operation)
        if address:
            host, port = address.split(":")
            ldfs = LiteDFSClient(host, port)
            if object == "file":
                if operation == "create":
                    try:
                        r = ldfs.create_file(args.local_path, args.remote_path, replica = args.replica)
                        if r:
                            print("create file[%s] success" % args.remote_path)
                        else:
                            print("create file[%s] failed" % args.remote_path)
                    except Exception as e:
                        print(e)
                elif operation == "delete":
                    if args.remote_path:
                        try:
                            r = ldfs.delete_file(args.remote_path)
                            if r:
                                print("delete file[%s] success" % args.remote_path)
                            else:
                                print("delete file[%s] failed" % args.remote_path)
                        except Exception as e:
                            print(e)
                elif operation == "move":
                    if args.source_path and args.target_path:
                        try:
                            r = ldfs.move_file(args.source_path, args.target_path)
                            if r:
                                print("move file[%s] to %s success" % (args.source_path, args.target_path))
                            else:
                                print("move file[%s] to %s failed" % (args.source_path, args.target_path))
                        except Exception as e:
                            print(e)
                elif operation == "rename":
                    if args.remote_path and args.new_name:
                        try:
                            r = ldfs.rename_file(args.remote_path, args.new_name)
                            if r:
                                print("rename file[%s] to %s success" % (args.remote_path, args.new_name))
                            else:
                                print("rename file[%s] to %s failed" % (args.remote_path, args.new_name))
                        except Exception as e:
                            print(e)
                elif operation == "update":
                    if args.remote_path and args.replica:
                        try:
                            r = ldfs.update_file(args.remote_path, args.replica)
                            if r:
                                print("update file[%s] success" % args.remote_path)
                            else:
                                print("update file[%s] failed" % args.remote_path)
                        except Exception as e:
                            print(e)
                elif operation == "download":
                    try:
                        r = ldfs.download_file(args.remote_path, args.local_path)
                        if r:
                            print("download file[%s => %s] success" % (args.remote_path, args.local_path))
                        else:
                            print("download file[%s => %s] failed" % (args.remote_path, args.local_path))
                    except Exception as e:
                        print(e)
                elif operation == "info":
                    if args.remote_path:
                        try:
                            r = ldfs.info_file(args.remote_path)
                            if r:
                                print(json.dumps(r, indent = 4, sort_keys = True))
                            else:
                                print("get file[%s]'s info failed" % args.remote_path)
                        except Exception as e:
                            print(e)
            elif object == "directory":
                if operation == "create":
                    if args.remote_path:
                        try:
                            r = ldfs.create_directory(args.remote_path)
                            if r:
                                print("create directory[%s] success" % args.remote_path)
                            else:
                                print("create directory[%s] failed" % args.remote_path)
                        except Exception as e:
                            print(e)
                elif operation == "delete":
                    if args.remote_path:
                        try:
                            r = ldfs.delete_directory(args.remote_path)
                            if r:
                                print("delete directory[%s] success" % args.remote_path)
                            else:
                                print("delete directory[%s] failed" % args.remote_path)
                        except Exception as e:
                            print(e)
                elif operation == "move":
                    if args.source_path and args.target_path:
                        try:
                            r = ldfs.move_directory(args.source_path, args.target_path)
                            if r:
                                print("move directory[%s] to %s success" % (args.source_path, args.target_path))
                            else:
                                print("move directory[%s] to %s failed" % (args.source_path, args.target_path))
                        except Exception as e:
                            print(e)
                elif operation == "rename":
                    if args.remote_path and args.new_name:
                        try:
                            r = ldfs.rename_directory(args.remote_path, args.new_name)
                            if r:
                                print("rename directory[%s] to %s success" % (args.remote_path, args.new_name))
                            else:
                                print("rename directory[%s] to %s failed" % (args.remote_path, args.new_name))
                        except Exception as e:
                            print(e)
                elif operation == "list":
                    if args.remote_path:
                        try:
                            r = ldfs.list_directory(args.remote_path, offset = args.offset, limit = args.limit, include_file = args.exclude_file, include_directory = args.exclude_directory)
                            if r:
                                print_table_result(
                                    r["children"],
                                    ["id", "type", "size", "name"],
                                    args
                                )
                            else:
                                print("list directory[%s] failed" % args.remote_path)
                        except Exception as e:
                            print(e)
            if object == "path":
                if operation == "info":
                    if args.remote_path:
                        try:
                            r = ldfs.info_path(args.remote_path)
                            if r:
                                print(json.dumps(r, indent = 4, sort_keys = True))
                            else:
                                print("get file[%s]'s info failed" % args.remote_path)
                        except Exception as e:
                            print(e)
            elif object == "cluster":
                if operation == "info":
                    r = requests.get(url)
                    if r.status_code == 200:
                        data = r.json()
                        if args.raw:
                            print(json.dumps(data, indent = 4, sort_keys = True))
                        else:
                            if data["result"] == "ok":
                                print("online nodes:")
                                print_table_result(
                                    data["info"]["online_nodes"],
                                    [
                                        "id",
                                        "node_id",
                                        "http_host",
                                        "http_port",
                                        "data_path",
                                    ],
                                    args
                                )
                                print("\noffline nodes:")
                                print_table_result(
                                    data["info"]["offline_nodes"],
                                    [
                                        "id",
                                        "node_id",
                                        "http_host",
                                        "http_port",
                                        "data_path",
                                    ],
                                    args
                                )
                            else:
                                print_table_result(
                                    [data],
                                    ["result", "message"],
                                    args
                                )
                    else:
                        print("error:\ncode: %s\ncontent: %s" % (r.status_code, r.content))
                        
    except Exception as e:
        logging.error(logging.traceback.format_exc())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import os
import sys
import json
import time
import logging

cwd = os.path.split(os.path.realpath(__file__))[0]
sys.path.insert(0, os.path.split(cwd)[0])

from litedfs.data.utils.common import disk_usage, size_pretty
from litedfs.data import logger

LOG = logging.getLogger(__name__)


if __name__ == "__main__":
    logger.config_logging(file_name = "test_disk_usage.log",
                          log_level = "DEBUG",
                          dir_name = os.path.join(cwd, "logs"),
                          day_rotate = False,
                          when = "D",
                          interval = 1,
                          max_size = 20,
                          backup_count = 5,
                          console = True)

    LOG.debug("test start")
    
    try:
        s = time.time()
        r = disk_usage("/home/breeze")
        ss = time.time()
        LOG.debug("%s, use %ss", r, ss - s)
        LOG.debug("free: %s", size_pretty(r["free"]))
    except Exception as e:
        LOG.exception(e)

    LOG.debug("test end")



# -*- coding: utf-8 -*-

import os
import json
import logging

import sqlalchemy
from sqlalchemy import func, exc
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, BigInteger, String, Text, Boolean, Date, DateTime, Numeric
from sqlalchemy.orm.exc import NoResultFound

from litedfs.name.utils import common
from litedfs.name.config import CONFIG

LOG = logging.getLogger(__name__)


BaseDataNodes = declarative_base()

class DataNodesTable(BaseDataNodes):
    __tablename__ = "data_nodes"

    id = Column(Integer, primary_key = True, autoincrement = True)
    node_id = Column(Text, unique = True, nullable = False, index = True)
    info = Column(Text, nullable = False)
    create_at = Column(DateTime, nullable = False, index = True)
    update_at = Column(DateTime, nullable = False, index = True)

    @classmethod
    def init_engine_and_session(cls):
        cls.engine = create_engine('sqlite:///' + os.path.join(CONFIG["data_path"], "data_nodes.db"), echo = False)
        cls.session = sessionmaker(bind = cls.engine)
        return cls.engine, cls.session

    def to_dict(self):
        return {
            "id": self.id,
            "node_id": self.node_id,
            "info": json.loads(self.info),
            "create_at": str(self.create_at), # "%Y-%m-%d %H:%M:%S.%f")
            "update_at": str(self.update_at),
        }

    def parse_dict(self, source):
        result = False

        attrs = [
            "node_id",
            "info",
            "create_at",
            "update_at",
        ]

        if hasattr(source, "__getitem__"):
            for attr in attrs:
                try:
                    setattr(self, attr, source[attr])
                except:
                    LOG.debug("some exception occured when extract %s attribute to object, i will discard it",
                        attr)
                    continue
            result = True
        else:
            LOG.debug("input param source does not have dict-like method, so i will do nothing at all!")
        return result

    def __repr__(self):
        return "application: %s" % self.to_dict()



# -*- coding: utf-8 -*-

import os
import time
import shutil
import logging
import threading
from threading import Thread
from queue import Queue, Empty

import tornado.ioloop
from tornado import gen

from litedfs.tool.viewer.handlers import storage
from litedfs.tool.viewer.utils.task_cache import TaskCache
from litedfs.tool.viewer.utils.common import joinpath, splitpath, listdir, Command
from litedfs.tool.viewer.config import CONFIG

LOG = logging.getLogger(__name__)

MessageQueue = Queue(0)


class MessageSender(object):
    _instance = None
    name = "message sender"

    def __new__(cls, interval = 0.1):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.interval = interval
            cls._instance.ioloop_service()
        return cls._instance

    @classmethod
    def instance(cls):
        return cls._instance

    def ioloop_service(self):
        self.periodic_message_sender = tornado.ioloop.PeriodicCallback(
            self.send_message, 
            self.interval * 1000
        )
        self.periodic_message_sender.start()

    @gen.coroutine
    def send_message(self):
        try:
            task = MessageQueue.get(block = False)
            handler, msg = task
            LOG.debug("send message: %s", msg)
            if isinstance(handler, str):
                if handler == "local":
                    storage.LocalSocketHandler.send_msgs(msg)
                elif handler == "remote":
                    storage.RemoteSocketHandler.send_msgs(msg)
            else:
                handler.write_message(msg)
        except Empty:
            pass
        except Exception as e:
            LOG.exception(e)

    def close(self):
        try:
            if self.periodic_message_sender:
                self.periodic_message_sender.stop()
            LOG.info("MessageSender close")
        except Exception as e:
            LOG.exception(e)


class StoppableThread(Thread):
    """
    Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition.
    """

    def __init__(self):
        super(StoppableThread, self).__init__()
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.isSet()


def download_directory(dir, remote_path, local_path, handler):
    source_path = os.path.join(remote_path, dir["name"])
    target_path = os.path.join(local_path, dir["name"])
    if not os.path.exists(target_path):
        os.mkdir(target_path)
    items, _ = handler.client.listdir(source_path, sort_by = "name", desc = False)
    for item in items:
        if item["type"] == "Directory":
            download_directory(item, source_path, target_path, handler)
        else:
            msg = {"cmd": "info"}
            try:
                source_file_path = os.path.join(source_path, item["name"])
                target_file_path = os.path.join(target_path, item["name"])
                if handler.client.download_file(source_file_path, target_file_path):
                    msg["info"] = "Download remote file [%s] to [%s] success" % (source_file_path, target_file_path)
                    LOG.info("Download remote file [%s] to [%s] success", source_file_path, target_file_path)
                else:
                    msg["cmd"] = "error"
                    msg["info"] = "Download remote file [%s] to [%s] failed" % (source_file_path, target_file_path)
                    LOG.info("Download remote file [%s] to [%s] failed", source_file_path, target_file_path)
                time.sleep(0.1)
            except Exception as e:
                LOG.exception(e)
                msg["cmd"] = "error"
                msg["info"] = str(e)
            MessageQueue.put([handler, msg])
    msg = {"cmd": "info"}
    msg["info"] = "Download remote directory [%s] to [%s] success" % (source_path, target_path)
    LOG.info("Download remote directory [%s] to [%s] success", source_path, target_path)
    MessageQueue.put([handler, msg])


def upload_directory(dir, local_path, remote_path, handler, replica = 1):
    source_path = os.path.join(local_path, dir["name"])
    target_path = os.path.join(remote_path, dir["name"])
    handler.client.mkdir(target_path)
    items, _ = listdir(source_path)
    for item in items:
        if item["type"] == "Directory":
            upload_directory(item, source_path, target_path, handler, replica = replica)
        else:
            msg = {"cmd": "info"}
            try:
                source_file_path = os.path.join(source_path, item["name"])
                target_file_path = os.path.join(target_path, item["name"])
                if handler.client.upload_file(source_file_path, target_file_path, replica = replica):
                    msg["info"] = "Upload file [%s] to [%s] success" % (source_file_path, target_file_path)
                    LOG.info("Upload file [%s] to [%s] success", source_file_path, target_file_path)
                else:
                    msg["cmd"] = "error"
                    msg["info"] = "Upload file [%s] to [%s] failed" % (source_file_path, target_file_path)
                    LOG.info("Upload file [%s] to [%s] failed", source_file_path, target_file_path)
                time.sleep(0.1)
            except Exception as e:
                LOG.exception(e)
                msg["cmd"] = "error"
                msg["info"] = str(e)
            MessageQueue.put([handler, msg])
    msg = {"cmd": "info"}
    msg["info"] = "Upload directory [%s] to [%s] success" % (source_path, target_path)
    LOG.info("Upload directory [%s] to [%s] success", source_path, target_path)
    MessageQueue.put([handler, msg])


def update_directory(dir, remote_path, handler, replica):
    dir_path = os.path.join(remote_path, dir["name"])
    items, _ = handler.client.listdir(dir_path, sort_by = "name", desc = False)
    for item in items:
        if item["type"] == "Directory":
            update_directory(item, dir_path, handler, replica = replica)
        else:
            msg = {"cmd": "info"}
            try:
                f_path = os.path.join(dir_path, item["name"])
                if handler.client.update_file(f_path, replica = replica):
                    msg["info"] = "Update remote file [%s] success" % f_path
                    LOG.info("Update remote file [%s] success", f_path)
                else:
                    msg["cmd"] = "error"
                    msg["info"] = "Update remote file [%s] failed" % f_path
                    LOG.info("Update remote file [%s] failed", f_path)
                time.sleep(0.1)
            except Exception as e:
                LOG.exception(e)
                msg["cmd"] = "error"
                msg["info"] = str(e)
            MessageQueue.put([handler, msg])
    msg = {"cmd": "info"}
    msg["info"] = "Update remote directory [%s] success" % dir_path
    LOG.info("Update remote directory [%s] success", dir_path)
    MessageQueue.put([handler, msg])


class TaskProcesser(StoppableThread):
    def __init__(self, pid):
        StoppableThread.__init__(self)
        Thread.__init__(self)
        self.pid = pid

    def run(self):
        LOG = logging.getLogger("task_processer")
        LOG.info("TaskProcesser(%03d) start", self.pid)
        try:
            while True:
                if not self.stopped():
                    try:
                        if not TaskCache.empty():
                            task = TaskCache.pop()
                            if task is not None:
                                if task["cmd"] == Command.delete:
                                    dir_path = joinpath(task["dir_path"])
                                    dirs = task["dirs"]
                                    files = task["files"]
                                    for d in dirs:
                                        msg = {}
                                        try:
                                            d_path = os.path.join(dir_path, d["name"])
                                            shutil.rmtree(d_path)
                                            msg["cmd"] = "info"
                                            msg["info"] = "Delete directory [%s] success" % d_path
                                        except Exception as e:
                                            LOG.exception(e)
                                            msg["cmd"] = "error"
                                            msg["info"] = str(e)
                                        MessageQueue.put([task["socket_handler"], msg])
                                    for f in files:
                                        msg = {}
                                        try:
                                            f_path = os.path.join(dir_path, f["name"])
                                            os.remove(f_path)
                                            msg["cmd"] = "info"
                                            msg["info"] = "Delete file [%s] success" % f_path
                                        except Exception as e:
                                            LOG.exception(e)
                                            msg["cmd"] = "error"
                                            msg["info"] = str(e)
                                        MessageQueue.put([task["socket_handler"], msg])
                                    msg = {"cmd": Command.need_refresh, "dir_path": dir_path}
                                    MessageQueue.put(["local", msg])
                                elif task["cmd"] == Command.paste:
                                    dir_path = joinpath(task["dir_path"])
                                    clipboard = task["clipboard"]
                                    dirs = clipboard["dirs"]
                                    files = clipboard["files"]
                                    for d in dirs:
                                        msg = {"cmd": "info"}
                                        try:
                                            source_path = os.path.join(clipboard["dir_path"], d["name"])
                                            target_path = os.path.join(dir_path, d["name"])
                                            if os.path.exists(target_path):
                                                msg["cmd"] = "warning"
                                                msg["info"] = "Directory [%s] already exists!" % target_path
                                                LOG.warning("Directory [%s] already exists!", target_path)
                                            elif not os.path.exists(source_path) or not os.path.isdir(source_path):
                                                msg["cmd"] = "warning"
                                                msg["info"] = "Directory [%s] doesn't exist!" % source_path
                                                LOG.warning("Directory [%s] doesn't exist!", source_path)
                                            else:
                                                if clipboard["type"] == Command.cut:
                                                    shutil.move(source_path, target_path)
                                                    msg["info"] = "Cut directory [%s] to [%s] success" % (source_path, target_path)
                                                    LOG.info("Cut directory [%s] to [%s] success", source_path, target_path)
                                                elif clipboard["type"] == Command.copy:
                                                    shutil.copytree(source_path, target_path)
                                                    msg["info"] = "Copy directory [%s] to [%s] success"%(source_path, target_path)
                                                    LOG.info("Copy directory [%s] to [%s] success", source_path, target_path)
                                                time.sleep(0.1)
                                                LOG.info("Paste directory from [%s] to [%s]", source_path, target_path)
                                        except Exception as e:
                                            LOG.exception(e)
                                            msg["cmd"] = "error"
                                            msg["info"] = str(e)
                                        MessageQueue.put([task["socket_handler"], msg])
                                    for f in files:
                                        msg = {"cmd": "info"}
                                        try:
                                            source_path = os.path.join(clipboard["dir_path"], f["name"])
                                            target_path = os.path.join(dir_path, f["name"])
                                            if os.path.exists(target_path):
                                                msg["cmd"] = "warning"
                                                msg["info"] = "File [%s] already exists!" % target_path
                                                LOG.warning("File [%s] already exists!", target_path)
                                            elif not os.path.exists(source_path) or not os.path.isfile(source_path):
                                                msg["cmd"] = "warning"
                                                msg["info"] = "File [%s] doesn't exists!" % source_path
                                                LOG.warning("File [%s] doesn't exists!", source_path)
                                            else:
                                                if clipboard["type"] == Command.cut:
                                                    shutil.move(source_path, target_path)
                                                    msg["info"] = "Cut file [%s] to [%s] success" % (source_path, target_path)
                                                    LOG.info("Cut file [%s] to [%s] success", source_path, target_path)
                                                elif clipboard["type"] == Command.copy:
                                                    shutil.copy(source_path, target_path)
                                                    msg["info"] = "Copy file [%s] to [%s] success" % (source_path, target_path)
                                                    LOG.info("Copy file [%s] to [%s] success", source_path, target_path)
                                                time.sleep(0.1)
                                                LOG.info("Paste file [%s] to [%s]", source_path, target_path)
                                        except Exception as e:
                                            LOG.exception(e)
                                            msg["cmd"] = "error"
                                            msg["info"] = str(e)
                                        MessageQueue.put([task["socket_handler"], msg])
                                    msg = {"cmd": Command.need_refresh, "dir_path": dir_path}
                                    MessageQueue.put(["local", msg])
                                elif task["cmd"] == Command.remote_delete:
                                    dir_path = joinpath(task["dir_path"])
                                    dirs = task["dirs"]
                                    files = task["files"]
                                    for d in dirs:
                                        msg = {}
                                        try:
                                            d_path = os.path.join(dir_path, d["name"])
                                            if task["socket_handler"].client.delete_directory(d_path):
                                                msg["cmd"] = "info"
                                                msg["info"] = "Delete remote directory [%s] success" % d_path
                                            else:
                                                msg["cmd"] = "error"
                                                msg["info"] = "Delete remote directory [%s] failed" % d_path
                                        except Exception as e:
                                            LOG.exception(e)
                                            msg["cmd"] = "error"
                                            msg["info"] = str(e)
                                        MessageQueue.put([task["socket_handler"], msg])
                                    for f in files:
                                        msg = {}
                                        try:
                                            f_path = os.path.join(dir_path, f["name"])
                                            if task["socket_handler"].client.delete_directory(f_path):
                                                msg["cmd"] = "info"
                                                msg["info"] = "Delete remote file [%s] success" % f_path
                                            else:
                                                msg["cmd"] = "error"
                                                msg["info"] = "Delete remote file [%s] failed" % f_path
                                        except Exception as e:
                                            LOG.exception(e)
                                            msg["cmd"] = "error"
                                            msg["info"] = str(e)
                                        MessageQueue.put([task["socket_handler"], msg])
                                    msg = {"cmd": Command.need_refresh, "dir_path": dir_path}
                                    MessageQueue.put(["remote", msg])
                                elif task["cmd"] == Command.remote_paste:
                                    dir_path = joinpath(task["dir_path"])
                                    clipboard = task["clipboard"]
                                    dirs = clipboard["dirs"]
                                    files = clipboard["files"]
                                    for d in dirs:
                                        msg = {"cmd": "info"}
                                        try:
                                            source_path = os.path.join(clipboard["dir_path"], d["name"])
                                            target_path = dir_path
                                            if task["socket_handler"].client.move_directory(source_path, target_path):
                                                msg["info"] = "Cut remote directory [%s] to [%s] success" % (source_path, target_path)
                                                LOG.info("Cut remote directory [%s] to [%s] success", source_path, target_path)
                                            else:
                                                msg["cmd"] = "error"
                                                msg["info"] = "Cut remote directory [%s] to [%s] failed" % (source_path, target_path)
                                                LOG.info("Cut remote directory [%s] to [%s] failed", source_path, target_path)
                                            time.sleep(0.1)
                                        except Exception as e:
                                            LOG.exception(e)
                                            msg["cmd"] = "error"
                                            msg["info"] = str(e)
                                        MessageQueue.put([task["socket_handler"], msg])
                                    for f in files:
                                        msg = {"cmd": "info"}
                                        try:
                                            source_path = os.path.join(clipboard["dir_path"], f["name"])
                                            target_path = dir_path
                                            if task["socket_handler"].client.move_file(source_path, target_path):
                                                msg["info"] = "Cut remote file [%s] to [%s] success" % (source_path, target_path)
                                                LOG.info("Cut remote file [%s] to [%s] success", source_path, target_path)
                                            else:
                                                msg["cmd"] = "error"
                                                msg["info"] = "Cut remote file [%s] to [%s] failed" % (source_path, target_path)
                                                LOG.info("Cut remote file [%s] to [%s] failed", source_path, target_path)
                                            time.sleep(0.1)
                                        except Exception as e:
                                            LOG.exception(e)
                                            msg["cmd"] = "error"
                                            msg["info"] = str(e)
                                        MessageQueue.put([task["socket_handler"], msg])
                                    msg = {"cmd": Command.need_refresh, "dir_path": dir_path}
                                    MessageQueue.put(["remote", msg])
                                elif task["cmd"] == Command.download:
                                    local_path = task["local_path"]
                                    remote_path = task["remote_path"]
                                    dirs = task["dirs"]
                                    files = task["files"]
                                    for d in dirs:
                                        download_directory(d, remote_path, local_path, task["socket_handler"])
                                    for f in files:
                                        msg = {"cmd": "info"}
                                        try:
                                            source_path = os.path.join(remote_path, f["name"])
                                            target_path = os.path.join(local_path, f["name"])
                                            if task["socket_handler"].client.download_file(source_path, target_path):
                                                msg["info"] = "Download remote file [%s] to [%s] success" % (source_path, target_path)
                                                LOG.info("Download remote file [%s] to [%s] success", source_path, target_path)
                                            else:
                                                msg["cmd"] = "error"
                                                msg["info"] = "Download remote file [%s] to [%s] failed" % (source_path, target_path)
                                                LOG.info("Download remote file [%s] to [%s] failed", source_path, target_path)
                                            time.sleep(0.1)
                                        except Exception as e:
                                            LOG.exception(e)
                                            msg["cmd"] = "error"
                                            msg["info"] = str(e)
                                        MessageQueue.put([task["socket_handler"], msg])
                                    msg = {"cmd": Command.need_refresh, "dir_path": local_path}
                                    MessageQueue.put(["local", msg])
                                elif task["cmd"] == Command.upload:
                                    local_path = task["local_path"]
                                    remote_path = task["remote_path"]
                                    dirs = task["dirs"]
                                    files = task["files"]
                                    replica = task["replica"] if "replica" in task else 1
                                    for d in dirs:
                                        upload_directory(d, local_path, remote_path, task["socket_handler"], replica = replica)
                                    for f in files:
                                        msg = {"cmd": "info"}
                                        try:
                                            source_path = os.path.join(local_path, f["name"])
                                            target_path = os.path.join(remote_path, f["name"])
                                            if task["socket_handler"].client.upload_file(source_path, target_path, replica = replica):
                                                msg["info"] = "Upload file [%s] to [%s] success" % (source_path, target_path)
                                                LOG.info("Upload file [%s] to [%s] success", source_path, target_path)
                                            else:
                                                msg["cmd"] = "error"
                                                msg["info"] = "Upload file [%s] to [%s] failed" % (source_path, target_path)
                                                LOG.info("Upload file [%s] to [%s] failed", source_path, target_path)
                                            time.sleep(0.1)
                                        except Exception as e:
                                            LOG.exception(e)
                                            msg["cmd"] = "error"
                                            msg["info"] = str(e)
                                        MessageQueue.put([task["socket_handler"], msg])
                                    msg = {"cmd": Command.need_refresh, "dir_path": remote_path}
                                    MessageQueue.put(["remote", msg])
                                elif task["cmd"] == Command.update:
                                    dir_path = joinpath(task["dir_path"])
                                    dirs = task["dirs"]
                                    files = task["files"]
                                    replica = task["replica"]
                                    for d in dirs:
                                        update_directory(d, dir_path, task["socket_handler"], replica = replica)
                                    for f in files:
                                        msg = {"cmd": "info"}
                                        try:
                                            f_path = os.path.join(dir_path, f["name"])
                                            if task["socket_handler"].client.update_file(f_path, replica = replica):
                                                msg["info"] = "Update remote file [%s] success" % f_path
                                                LOG.info("Update remote file [%s] success", f_path)
                                            else:
                                                msg["cmd"] = "error"
                                                msg["info"] = "Update remote file [%s] failed" % f_path
                                                LOG.info("Update remote file [%s] failed", f_path)
                                            time.sleep(0.1)
                                        except Exception as e:
                                            LOG.exception(e)
                                            msg["cmd"] = "error"
                                            msg["info"] = str(e)
                                        MessageQueue.put([task["socket_handler"], msg])
                                    msg = {"cmd": Command.need_refresh, "dir_path": dir_path}
                                    MessageQueue.put(["remote", msg])
                                elif task["cmd"] == Command.preview:
                                    dir_path = joinpath(task["dir_path"])
                                    file = task["file"]
                                    file_type = file["type"].lower()
                                    file_path = os.path.join(dir_path, file["name"])
                                    msg = {}
                                    if file_type == ".zip":
                                        namelist = task["socket_handler"].client.preview_zip_file(file_path)
                                        if namelist:
                                            msg["cmd"] = Command.preview
                                            msg["file_path"] = file_path
                                            msg["data"] = namelist
                                            msg["type"] = file_type
                                        else:
                                            msg["cmd"] = "error"
                                            msg["info"] = "preview file [%s] failed" % file_path
                                    elif file["name"].lower() == ".gitignore" or file_type in [".txt", ".log", ".yml", ".json", ".md", ".py", ".c", ".xml", ".go", ".sh", ".html"]:
                                        content = task["socket_handler"].client.preview_text_file(file_path)
                                        if content:
                                            msg["cmd"] = Command.preview
                                            msg["file_path"] = file_path
                                            msg["data"] = content
                                            msg["type"] = file_type
                                        else:
                                            msg["cmd"] = "error"
                                            msg["info"] = "preview file [%s] failed" % file_path
                                    else:
                                        msg["cmd"] = "error"
                                        msg["info"] = "preview does not support this type of file"
                                    MessageQueue.put([task["socket_handler"], msg])
                            else:
                                time.sleep(0.5)
                            LOG.info("TaskProcesser(%03d) process task: %s", self.pid, task)
                        else:
                            time.sleep(0.5)
                    except Exception as e:
                        LOG.exception(e)
                else:
                    LOG.info("TaskProcesser(%03d) exit by siginal", self.pid)
                    break
        except Exception as e:
            LOG.exception(e)
        LOG.info("TaskProcesser(%03d) exit", self.pid)



# -*- coding: utf-8 -*-

import logging

LOG = logging.getLogger(__name__)


class TaskCache(object):
    cache = []
    max_size = 10

    @classmethod
    def set_max_size(cls, size):
        cls.max_size = size

    @classmethod
    def push(cls, task):
        cls.cache.append(task)

    @classmethod
    def pop(cls):
        result = None
        try:
            result = cls.cache.pop(0)
        except IndexError:
            pass
        except Exception as e:
            LOG.exception(e)
        return result

    @classmethod
    def empty(cls):
        return len(cls.cache) == 0

    @classmethod
    def full(cls):
        return len(cls.cache) >= cls.max_size

    @classmethod
    def size(cls):
        return len(cls.cache)


# -*- coding: utf-8 -*-

import os
import logging
import argparse

from litedfs.version import __version__
from litedfs.name.utils.fs_core import F, C
from litedfs.name.utils.append_log import AppendLogJson
from litedfs.name.config import CONFIG, load_config
from litedfs.name import logger

LOG = logging.getLogger(__name__)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog = 'migrate_fsimage.py')
    parser.add_argument("-c", "--config", required = True, help = "configuration file path")
    parser.add_argument("-v", "--version", action = 'version', version = '%(prog)s ' + __version__)
    args = parser.parse_args()
    if args.config:
        success = load_config(args.config)
        if success:
            logger.config_logging(file_name = "migrate_fsimage.log",
                                  log_level = CONFIG["log_level"],
                                  dir_name = CONFIG["log_path"],
                                  day_rotate = False,
                                  when = "D",
                                  interval = 1,
                                  max_size = 20,
                                  backup_count = 5,
                                  console = True)

            LOG.info("migrate start")
            
            fsimage_new = AppendLogJson(os.path.join(CONFIG["data_path"], "fsimage.new"))
            fsimage = AppendLogJson(os.path.join(CONFIG["data_path"], "fsimage"))
            for line in fsimage.iterlines():
                if line[F.cmd] == C.makedir:
                    line[F.cmd] = C.makedirs
                fsimage_new.writeline(line)
            fsimage_new.close()
            fsimage.close()

            editlog_new = AppendLogJson(os.path.join(CONFIG["data_path"], "editlog.new"))
            editlog = AppendLogJson(os.path.join(CONFIG["data_path"], "editlog"))
            for line in editlog.iterlines():
                if line[F.cmd] == C.makedir:
                    line[F.cmd] = C.makedirs
                editlog_new.writeline(line)
            editlog_new.close()
            editlog.close()

            LOG.info("migrate end")
        else:
            print("failed to load configuration: %s" % args.config)
    else:
        parser.print_help()


# -*- coding: utf-8 -*-

import os
import time
import json
import logging
import datetime
import zipfile

from litedfs_client.client import LiteDFSClient
from litedfs.tool.viewer.utils.common import joinpath, splitpath, listsort, sha1sum
from litedfs.tool.viewer.config import CONFIG

LOG = logging.getLogger(__name__)


class RemoteStorage(object):
    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.client = LiteDFSClient(self.host, self.port)

    def listdir(self, dir_path, sort_by = "name", desc = False, offset = 0, limit = -1):
        dirs = []
        files = []
        try:
            r = self.client.list_directory(dir_path)
            if r:
                if "result" in r and r["result"] == "ok":
                    n = 1
                    for c in r["children"]:
                        if c["type"] == "directory":
                            d_path = os.path.join(dir_path, c["name"])
                            dirs.append({
                                "num": n,
                                "name": c["name"],
                                "sha1": sha1sum(d_path),
                                "type": "Directory",
                                "size": c["size"],
                                "ctime": "",
                                "mtime": ""
                            })
                        elif c["type"] == "file":
                            f_path = os.path.join(dir_path, c["name"])
                            f = {
                                "num": n,
                                "name": c["name"],
                                "sha1": sha1sum(f_path),
                                "type": os.path.splitext(c["name"])[-1],
                                "size": c["size"],
                                "ctime": "",
                                "mtime": ""
                            }
                            if "current_replica" in c:
                                f["current_replica"] = c["current_replica"]
                            if "replica" in c:
                                f["replica"] = c["replica"]
                            files.append(f)
                        n += 1
                    items, total = listsort(dirs, files, sort_by = sort_by, desc = desc, offset = offset, limit = limit)
        except Exception as e:
            LOG.exception(e)
        return items, total

    def list_storage(self, home_path, dir_path, sort_by = "name", desc = False, offset = 0, limit = -1):
        data = {}
        try:
            r = self.client.list_directory(dir_path, offset = offset, limit = limit)
            if r:
                if "result" in r and r["result"] == "ok":
                    dirs = []
                    files = []
                    total = r["total"]
                    n = 1
                    for c in r["children"]:
                        if c["type"] == "directory":
                            d_path = os.path.join(dir_path, c["name"])
                            dirs.append({
                                "num": n,
                                "name": c["name"],
                                "sha1": sha1sum(d_path),
                                "type": "Directory",
                                "size": c["size"],
                                "ctime": datetime.datetime.fromtimestamp(c["ctime"]).strftime("%Y-%m-%d %H:%M:%S") if "ctime" in c and c["ctime"] else "",
                                "mtime": datetime.datetime.fromtimestamp(c["mtime"]).strftime("%Y-%m-%d %H:%M:%S") if "mtime" in c and c["mtime"] else ""
                            })
                        elif c["type"] == "file":
                            f_path = os.path.join(dir_path, c["name"])
                            f = {
                                "num": n,
                                "name": c["name"],
                                "sha1": sha1sum(f_path),
                                "type": os.path.splitext(c["name"])[-1],
                                "size": c["size"],
                                "ctime": datetime.datetime.fromtimestamp(c["ctime"]).strftime("%Y-%m-%d %H:%M:%S") if "ctime" in c and c["ctime"] else "",
                                "mtime": datetime.datetime.fromtimestamp(c["mtime"]).strftime("%Y-%m-%d %H:%M:%S") if "mtime" in c and c["mtime"] else ""
                            }
                            if "current_replica" in c:
                                f["current_replica"] = c["current_replica"]
                            if "replica" in c:
                                f["replica"] = c["replica"]
                            files.append(f)
                        n += 1
                    items, _ = listsort(dirs, files, sort_by = sort_by, desc = desc, n_start = offset + 1)
                    data["items"] = items
                    data["offset"] = offset
                    data["limit"] = limit
                    data["total"] = total
                    data["sort"] = {"name": sort_by, "desc": desc}
                    data["dir_path"] = splitpath(dir_path)
                    data["home_path"] = splitpath(home_path)
                    data["home_path_string"] = home_path
        except Exception as e:
            LOG.exception(e)
        return data

    def rename(self, dir_path, new_name):
        result = False
        try:
            result = self.client.rename_file(dir_path, new_name)
        except Exception as e:
            LOG.exception(e)
        return result

    def mkdir(self, dir_path):
        result = False
        try:
            result = self.client.create_directory(dir_path)
        except Exception as e:
            LOG.exception(e)
        return result

    def delete_file(self, file_path):
        result = False
        try:
            result = self.client.delete_file(file_path)
        except Exception as e:
            LOG.exception(e)
        return result

    def delete_directory(self, dir_path):
        result = False
        try:
            result = self.client.delete_directory(dir_path)
        except Exception as e:
            LOG.exception(e)
        return result

    def move_file(self, source_path, target_path):
        result = False
        try:
            result = self.client.move_file(source_path, target_path)
        except Exception as e:
            LOG.exception(e)
        return result

    def move_directory(self, source_path, target_path):
        result = False
        try:
            result = self.client.move_directory(source_path, target_path)
        except Exception as e:
            LOG.exception(e)
        return result

    def download_file(self, source_path, target_path):
        result = False
        try:
            result = self.client.download_file(source_path, target_path)
        except Exception as e:
            LOG.exception(e)
        return result

    def upload_file(self, source_path, target_path, replica = 1):
        result = False
        try:
            result = self.client.create_file(source_path, target_path, replica = replica)
        except Exception as e:
            LOG.exception(e)
        return result

    def update_file(self, file_path, replica = 1):
        result = False
        try:
            result = self.client.update_file(file_path, replica)
        except Exception as e:
            LOG.exception(e)
        return result

    def preview_zip_file(self, file_path):
        result = False
        try:
            fp = self.client.open_remote_file(file_path)
            if fp:
                z = zipfile.ZipFile(fp)
                result = z.namelist()
        except Exception as e:
            LOG.exception(e)
        return result

    def preview_text_file(self, file_path):
        result = False
        try:
            fp = self.client.open_remote_file(file_path)
            if fp:
                result = fp.read().decode()
        except Exception as e:
            LOG.exception(e)
        return result



# -*- coding: utf-8 -*-

import os
import time
import json
import hashlib
import logging
import datetime

from tornado import ioloop
from tornado import gen
import psutil

from litedfs.tool.viewer.config import CONFIG

LOG = logging.getLogger(__name__)

BUF_SIZE = 65536


class Servers(object):
    HTTP_SERVER = None
    SERVERS = []
    TORNADO_INSTANCE = None


async def shutdown():
    LOG.info("Stopping Service(%s:%s)", CONFIG["http_host"], CONFIG["http_port"])
    if Servers.HTTP_SERVER:
        Servers.HTTP_SERVER.stop()
        LOG.info("Stop http server!")
    for s in Servers.SERVERS:
        if hasattr(s, "close"):
            s.close()
        elif hasattr(s, "stop"):
            s.stop()
        if hasattr(s, "name"):
            LOG.info("Stop %s server!", s.name)
        else:
            LOG.info("Stop nameless server!")
    await gen.sleep(1)
    LOG.info("Will shutdown ...")
    ioloop.IOLoop.current().stop()


def sig_handler(sig, frame):
    LOG.warning("sig_handler Caught signal: %s", sig)
    ioloop.IOLoop.current().add_callback_from_signal(shutdown)


class Errors(object):
    OK = "ok"
    errors = {
        "ServerException": {"name": "ServerException", "message": "server exception"},
        "InvalidParameters": {"name": "InvalidParameters", "message": "invalid parameters"},
        "OperationFailed": {"name": "OperationFailed", "message": "operation failed"},
        "AppNotExists": {"name": "AppNotExists", "message": "application not exists"},
        "TaskNotExists": {"name": "TaskNotExists", "message": "task not exists"},
        "OperationRunning": {"name": "OperationRunning", "message": "operation running"},
        "WorkspaceNotExists": {"name": "WorkspaceNotExists", "message": "workspace not exists"},
        "WorkspaceNotPacked": {"name": "WorkspaceNotPacked", "message": "workspace not packed"},
    }

    @classmethod
    def set_result_error(cls, error_name, result, message = ""):
        if error_name in cls.errors:
            result["result"] = error_name
            if message == "":
                result["message"] = cls.errors[error_name]["message"]
            else:
                result["message"] = message
        else:
            result["result"] = "UnknownError"
            result["message"] = "unknown error"


class Stage(object):
    pending = "pending"
    running = "running"
    finished = "finished"


class Status(object):
    fail = "fail"
    success = "success"
    kill = "kill"
    cancel = "cancel"
    terminate = "terminate"


class Signal(object):
    kill = -9
    terminate = -15
    stop = -15
    cancel = -50


class JSONLoadError(Exception):
    def __init__(self, message):
        self.message = message


class MetaNotDictError(Exception):
    def __init__(self, message):
        self.message = message


def sha1sum(content):
    sha1 = hashlib.sha1()
    sha1.update(content.encode("utf-8"))
    return sha1.hexdigest()


def file_sha1sum(file_path):
    sha1 = hashlib.sha1()
    with open(file_path, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha1.update(data)
    return sha1.hexdigest()


def file_md5sum(file_path):
    md5 = hashlib.md5()
    with open(file_path, 'rb') as fp:
        while True:
            data = fp.read(BUF_SIZE)
            if not data:
                break
            md5.update(data)
    return md5.hexdigest()


class Command(object):
    cd = "cd"
    refresh = "refresh"
    rename = "rename"
    mkdir = "mkdir"
    delete = "delete"
    copy = "copy"
    cut = "cut"
    paste = "paste"
    remote_delete = "remote_delete"
    remote_paste = "remote_paste"
    download = "download"
    upload = "upload"
    need_refresh = "need_refresh"
    update = "update"
    preview = "preview"
    change_page = "change_page"


def splitall(path):
    allparts = []
    while True:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts


def get_workspace_path(create_at, task_id = None, action_name = None):
    result = ""
    date_create_at = datetime.datetime.strptime(create_at, "%Y-%m-%d %H:%M:%S.%f")
    date_directory_name = date_create_at.strftime("%Y-%m-%d")
    if task_id and action_name:
        result = os.path.join(CONFIG["data_path"], "tmp", "workspace", date_directory_name, task_id[:2], task_id[2:4], task_id, action_name)
    elif task_id:
        result = os.path.join(CONFIG["data_path"], "tmp", "workspace", date_directory_name, task_id[:2], task_id[2:4], task_id)
    else:
        result = os.path.join(CONFIG["data_path"], "tmp", "workspace", date_directory_name)
    return result


def init_storage():
    directories = [
        os.path.join(CONFIG["data_path"], "applications"),
        os.path.join(CONFIG["data_path"], "tmp", "workspace"),
        os.path.join(CONFIG["data_path"], "tmp", "download"),
    ]
    for d in directories:
        if not os.path.exists(d) or not os.path.isdir(d):
            os.makedirs(d)


def get_file_size(size):
    result = ""
    try:
        if size > 1024*1014*1024:
            result = "%.3f G"%(size/1024.0/1024.0/1024.0)
        elif size > 1024*1024:
            result = "%.3f M"%(size/1024.0/1024.0)
        elif size > 1024:
            result = "%.3f K"%(size/1024.0)
        else:
            result = "%d B"%size
    except Exception as e:
        LOG.exception(e)
        result = "0 B"
    return result


def makekey(c):
    if isinstance(c, int):
        return c
    elif isinstance(c, str):
        return c.lower()


def listsort(dirs, files, sort_by = "name", desc = False, offset = 0, limit = -1, n_start = 1):
    dirs_keys = []
    dirs_tree = {}
    dirs_sort = []
    files_keys = []
    files_tree = {}
    files_sort = []
    result = []
    for d in dirs:
        dirs_keys.append(d[sort_by])
        if d[sort_by] in dirs_tree:
            dirs_tree[d[sort_by]].append(d)
        else:
            dirs_tree[d[sort_by]] = []
            dirs_tree[d[sort_by]].append(d)
    dirs_keys = list(set(dirs_keys))
    dirs_keys.sort(key = makekey, reverse = desc)
    # LOG.info("Dirs_keys: %s", dirs_keys)
    n = n_start
    for k in dirs_keys:
        for d in dirs_tree[k]:
            d["num"] = n
            d["size"] = get_file_size(d["size"])
            dirs_sort.append(d)
            n += 1
    for f in files:
        files_keys.append(f[sort_by])
        if f[sort_by] in files_tree:
            files_tree[f[sort_by]].append(f)
        else:
            files_tree[f[sort_by]] = []
            files_tree[f[sort_by]].append(f)
    files_keys = list(set(files_keys))
    files_keys.sort(key = makekey, reverse = desc)
    # LOG.info("Files_keys: %s", files_keys)
    for k in files_keys:
        for f in files_tree[k]:
            f["num"] = n
            f["size"] = get_file_size(f["size"])
            files_sort.append(f)
            n += 1
    items = []
    items.extend(dirs_sort)
    items.extend(files_sort)
    total_len = len(items)
    if limit == -1:
        result = items
    else:
        result = items[offset:offset + limit]

    return (result, total_len)


def listdir(dir_path = ".", sort_by = "name", desc = False, offset = 0, limit = -1):
    dirs = []
    files = []
    try:
        dirs_list = [d for d in os.listdir(dir_path) if os.path.isdir(os.path.join(dir_path, d))]
        files_list = [f for f in os.listdir(dir_path) if os.path.isfile(os.path.join(dir_path, f))]
        dirs_list.sort()
        files_list.sort()
        n = 1
        for d in dirs_list:
            d_path = os.path.join(dir_path, d)
            dirs.append({
                "num":n,
                "name":d,
                "sha1":sha1sum(d_path),
                "type":"Directory",
                "size":os.path.getsize(d_path),
                "ctime":time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(os.path.getctime(d_path))),
                "mtime":time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(os.path.getmtime(d_path)))
            })
            n += 1
        for f in files_list:
            f_path = os.path.join(dir_path, f)
            files.append({
                "num":n,
                "name":f,
                "sha1":sha1sum(f_path),
                "type":os.path.splitext(f)[-1],
                "size":os.path.getsize(f_path),
                "ctime":time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(os.path.getctime(f_path))),
                "mtime":time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(os.path.getmtime(f_path)))
            })
            n += 1
    except Exception as e:
        LOG.exception(e)
    return listsort(dirs, files, sort_by = sort_by, desc = desc, offset = offset, limit = limit)


def joinpath(dir_list):
    dir_path = dir_list[0]
    dir_list = dir_list[1:]
    for d in dir_list:
        dir_path = os.path.join(dir_path, d)
    return dir_path


def splitpath(dir_path):
    dir_list = []
    dir_path, dir_last = os.path.split(dir_path)
    while dir_last != "":
        dir_list.append(dir_last)
        dir_path, dir_last = os.path.split(dir_path)
    dir_list.append(dir_path)
    dir_list.reverse()
    return dir_list


def list_storage(home_path, dir_path, sort_by = "name", desc = False, offset = 0, limit = -1):
    disk_usage = psutil.disk_usage(dir_path)
    disk_partitions = psutil.disk_partitions()
    data = {}
    items, total = listdir(dir_path = dir_path, sort_by = sort_by, desc = desc, offset = offset, limit = limit)
    data["items"] = items
    data["offset"] = offset
    data["limit"] = limit
    data["total"] = total
    data["sort"] = {"name":sort_by, "desc":desc}
    data["dir_path"] = splitpath(dir_path)
    data["home_path"] = splitpath(home_path)
    data["home_path_string"] = home_path
    data["disk_usage"] = {
        "total":get_file_size(disk_usage.total),
        "used":get_file_size(disk_usage.used),
        "free":get_file_size(disk_usage.free),
        "percent":disk_usage.percent
    }
    data["disk_partitions"] = [{"mountpoint":splitpath(p.mountpoint), "device":p.device} for p in disk_partitions]
    p_mountpoint_length = 0
    for n, p in enumerate(data["disk_partitions"]):
        mountpoint_path = joinpath(p["mountpoint"])
        if mountpoint_path in dir_path and len(mountpoint_path) > p_mountpoint_length:
            p_mountpoint_length = len(mountpoint_path)
            data["current_partition"] = n
    return data



# -*- coding: utf-8 -*-

import json
import time
import logging

from tornado import web
from tornado import gen

from litedfs.tool.viewer.handlers.base import BaseHandler, BaseSocketHandler
from litedfs.tool.viewer.config import CONFIG

LOG = logging.getLogger("__name__")


class ClusterHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        self.render(
            "cluster/cluster.html",
            current_nav = "cluster",
            manager_host = "%s:%s" % (
                self.get_name_http_host(),
                CONFIG["name_http_port"])
        )



# -*- coding: utf-8 -*-

import os
import json
import logging

from litedfs.name.config import CONFIG

LOG = logging.getLogger(__name__)


class AppendLog(object):
    def __init__(self, log_path):
        self.log_path = log_path
        self.log_file = open(self.log_path, "a")

    def writeline(self, line):
        result = False
        try:
            line += "\n"
            self.log_file.write(line)
            self.log_file.flush()
            result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def iterlines(self):
        fp = open(self.log_path, "r")
        line = fp.readline().strip()
        while line:
            yield line
            line = fp.readline().strip()
        fp.close()

    def close(self):
        try:
            if self.log_file:
                self.log_file.close()
        except Exception as e:
            LOG.exception(e)


class AppendLogJson(AppendLog):
    def writeline(self, data = {}):
        result = False
        try:
            line = json.dumps(data, separators = (",", ":")) + "\n"
            self.log_file.write(line)
            self.log_file.flush()
            result = True
        except Exception as e:
            LOG.exception(e)
        return result

    def iterlines(self):
        fp = open(self.log_path, "r")
        line = fp.readline().strip()
        while line:
            yield json.loads(line)
            line = fp.readline().strip()
        fp.close()



# -*- coding: utf-8 -*-

import os
import json
import time
import random
import logging
from copy import deepcopy

from tornado import gen, ioloop

from litedfs.name.utils.append_log import AppendLogJson
from litedfs.name.utils.listener import Connection
from litedfs.name.utils.common import file_sha1sum, file_md5sum, Errors, splitall
from litedfs.name.config import CONFIG

LOG = logging.getLogger(__name__)


class F(object):
    children = "c"
    type = "t"
    file = "f"
    dir = "d"
    info = "i"
    cmd = "c"
    new_name = "n"
    path = "p"
    source_path = "s"
    target_path = "t"
    id = "id"
    replica = "r"


class C(object):
    create = "c"
    makedirs = "mds"
    makedir = "md"
    delete = "d"
    rename = "r"
    move = "m"
    copy = "cp"
    update_replica = "ur"
    update_file_info = "ufi"
    update_parent_dirs = "upd"


class InvalidValueError(Exception):
    def __init__(self, message):
        self.message = message


class SameNameExistsError(Exception):
    def __init__(self, message):
        self.message = message


class SameNameFileExistsError(Exception):
    def __init__(self, message):
        self.message = message


class FileNotExistsError(Exception):
    def __init__(self, message):
        self.message = message


class ParentDirectoryNotExistsError(Exception):
    def __init__(self, message):
        self.message = message


class TargetPathMustDirectoryError(Exception):
    def __init__(self, message):
        self.message = message


class SourcePathNotExistsError(Exception):
    def __init__(self, message):
        self.message = message


class TargetPathNotExistsError(Exception):
    def __init__(self, message):
        self.message = message


class RecoverFailedError(Exception):
    def __init__(self, message):
        self.message = message


class FileSystemTree(object):
    _instance = None
    name = "file_system_tree"

    def __new__(cls, interval = 10):
        if not cls._instance:
            cls._instance = object.__new__(cls)
            cls._instance.tree = {F.children: {}, F.type: "root"}
            cls._instance.files = {}
            cls._instance.editlog = None
            cls._instance.status = "booting"
            cls._instance.locks = {}
            cls._instance.interval = interval
            cls._instance.ioloop_service()
        return cls._instance

    @classmethod
    def instance(cls):
        if cls._instance and cls._instance.status == "ready":
            return cls._instance
        else:
            return None

    def ioloop_service(self):
        self.periodic_lock_service = ioloop.PeriodicCallback(
            self.check_file_lock, 
            self.interval * 1000
        )
        self.periodic_lock_service.start()

    @gen.coroutine
    def recover(self):
        self.status = "recovering"
        success = yield self.load_fsimage()
        if success:
            success = yield self.load_editlog()
            if success:
                success = yield self.dump_fsimage()
                if success:
                    self.editlog = AppendLogJson(os.path.join(CONFIG["data_path"], "editlog"))
                    # TODO: synchronize between name node and data nodes
                    self.status = "ready"
                else:
                    raise RecoverFailedError("dump fsimage failed")
            else:
                raise RecoverFailedError("load editlog failed")
        else:
            raise RecoverFailedError("load fsimage failed")

    def set_file_lock(self, file_path, ttl = 60):
        result = False
        if file_path not in self.locks:
            self.locks[file_path] = time.time() + ttl
            result = True
        return result

    def unset_file_lock(self, file_path):
        try:
            if file_path in self.locks:
                del self.locks[file_path]
        except Exception as e:
            LOG.exception(e)

    def update_file_lock(self, file_path, ttl = 60):
        result = False
        if file_path in self.locks:
            self.locks[file_path] = time.time() + ttl
            result = True
        return result

    @gen.coroutine
    def check_file_lock(self):
        try:
            now = time.time()
            files = list(self.locks.keys())
            for f in files:
                if self.locks[f] < now:
                    del self.locks[f]
                    LOG.info("release write file lock: %s", f)
        except Exception as e:
            LOG.exception(e)

    def create(self, file_path, file_info, recover = False):
        result = False
        dir_path, file_name = os.path.split(file_path)
        parent = self.makedirs(dir_path)
        if parent:
            file_id = file_info["id"]
            if file_name in parent[F.children]:
                raise SameNameExistsError("same file name exists: %s" % file_name)
            elif file_name == "":
                raise InvalidValueError("file name can't be empty string: %s" % file_name)
            else:
                parent[F.children][file_name] = {F.type: F.file, F.id: file_id}
                self.files[file_id] = file_info
                if self.editlog:
                    self.editlog.writeline({F.cmd: C.create, F.path: file_path, F.info: file_info})
                if not recover:
                    self.update_parent_dirs(dir_path, file_info)
            result = True
        return result

    def delete(self, file_path, recover = False):
        result = True
        dir_path, name = os.path.split(file_path)
        exists, file_type, file, parent = self.get_info(file_path)
        if exists:
            del parent[F.children][name]
            if file[F.type] == F.file:
                file_id = file[F.id]
                del self.files[file_id]
                task = {"command": "delete", "name": file_id}
                for i in Connection.id_decompress:
                    Connection.push_task(i, task)
                if not recover:
                    self.update_parent_dirs(dir_path)
            elif file[F.type] == F.dir:
                self.delete_files(file)
                if not recover:
                    self.update_parent_dirs(dir_path)
            if self.editlog:
                self.editlog.writeline({F.cmd: C.delete, F.path: file_path})
        return result

    @gen.coroutine
    def delete_files(self, file):
        if F.children in file and file[F.children]:
            for name in file[F.children]:
                child = file[F.children][name]
                yield self.delete_files(child)
        else:
            if file[F.type] == F.file:
                file_id = file[F.id]
                del self.files[file_id]
                task = {"command": "delete", "name": file_id}
                for i in Connection.id_decompress:
                    Connection.push_task(i, task)
                LOG.debug("delete file: %s", file)
                yield gen.moment

    def get_file_info(self, file_path):
        result = False
        exists, file_type, file, _ = self.get_info(file_path)
        if exists:
            file_id = file[F.id]
            result = self.files[file_id]
        return result

    def rename(self, file_path, new_name, recover = False):
        result = True
        if self.is_valid_name(new_name):
            dir_path, name = os.path.split(file_path)
            exists, file_type, file, parent = self.get_info(file_path)
            if exists:
                if new_name not in parent[F.children]:
                    now = int(time.time())
                    parent[F.children][new_name] = file
                    del parent[F.children][name]
                    if self.editlog:
                        self.editlog.writeline({F.cmd: C.rename, F.path: file_path, F.new_name: new_name})
                    if file_type == F.file:
                        file_id = file[F.id]
                        file_info = self.files[file_id]
                        file_info["mtime"] = now
                        self.update_file_info(os.path.join(dir_path, new_name), file_info, recover = recover)
                    elif file_type == F.dir:
                        file[F.info]["mtime"] = now
                        if not recover:
                            self.update_parent_dirs(os.path.join(dir_path, new_name), file[F.info])
                else:
                    raise SameNameExistsError("same file name exists: %s" % new_name)
            else:
                raise FileNotExistsError("file not exists: %s" % file_path)
        else:
            raise InvalidValueError("invailed file name: %s" % new_name)
        return result

    def update_file_info(self, file_path, file_info, recover = False):
        result = False
        dir_path, _ = os.path.split(file_path)
        exists, file_type, file, parent = self.get_info(file_path)
        if exists:
            if file_type == F.file:
                file_id = file[F.id]
                self.files[file_id] = file_info
                if self.editlog:
                    self.editlog.writeline({F.cmd: C.update_file_info, F.path: file_path, F.info: file_info})
                if not recover:
                    self.update_parent_dirs(dir_path, file_info)
                result = True
            else:
                raise InvalidValueError("must by file not directory: %s" % file_path)
        else:
            raise FileNotExistsError("file not exists: %s" % file_path)
        return result

    @gen.coroutine
    def update_replica(self, file_path, replica):
        result = False
        exists, file_type, file, parent = self.get_info(file_path)
        if exists:
            now = int(time.time())
            if file_type == F.file:
                file_id = file[F.id]
                file_info = self.files[file_id]
                file_info["replica"] = replica
                data_nodes = Connection.get_node_infos()
                data_node_ids = list(data_nodes.keys())
                if (
                        (file_info["replica"] > file_info["current_replica"] and len(data_nodes) >= file_info["current_replica"]) or
                        (file_info["replica"] < file_info["current_replica"] and len(data_nodes) >= file_info["replica"])
                    ):
                    for block in file_info["blocks"]:
                        old_node_ids = []
                        new_node_ids = []
                        for node_id in data_node_ids:
                            if node_id not in block[2]:
                                if not data_nodes[node_id][2]:
                                    new_node_ids.append(node_id)
                            else:
                                old_node_ids.append(node_id)
                        if old_node_ids:
                            delta = file_info["replica"] - len(old_node_ids)
                            block[2] = []
                            block[2].extend(old_node_ids)
                            if delta > 0: # increase block replica
                                if new_node_ids:
                                    if len(new_node_ids) <= delta:
                                        block[2].extend(new_node_ids)
                                    else:
                                        new_node_ids = random.sample(new_node_ids, delta)
                                        block[2].extend(new_node_ids)
                                    source_node_id = random.choice(old_node_ids)
                                    task = {"command": "replicate", "name": file_id, "block": block[0], "ids": new_node_ids}
                                    Connection.push_task(source_node_id, task)
                                else:
                                    LOG.warning("can not increase block replica, no usable data node")
                            elif delta < 0: # decrease block replica
                                delta = -delta
                                if len(block[2]) > delta:
                                    delete_node_ids = random.sample(block[2], delta)
                                    block[2] = [i for i in block[2] if i not in delete_node_ids]
                                    for delete_node_id in delete_node_ids:
                                        task = {"command": "delete", "name": file_id, "block": block[0]}
                                        Connection.push_task(delete_node_id, task)
                        yield gen.moment
                    file_info["current_replica"] = replica
                    for block in file_info["blocks"]:
                        if len(block[2]) < file_info["current_replica"]:
                            file_info["current_replica"] = len(block[2])
                    if not recover:
                        file_info["mtime"] = now
                result = True
                if self.editlog:
                    self.editlog.writeline({F.cmd: C.update_file_info, F.path: file_path, F.info: file_info})
            else:
                raise InvalidValueError("must by file not directory: %s" % file_path)
        else:
            raise FileNotExistsError("file not exists: %s" % file_path)
        return result

    def is_valid_name(self, name):
        result = True
        if "/" in name:
            result = False
        return result

    def move(self, source_path, target_path):
        result = False
        _, name = os.path.split(source_path)
        source_exists, _, source_file, source_parent = self.get_info(source_path)
        if source_exists:
            target_exists, target_type, target_file, _ = self.get_info(target_path)
            if target_exists:
                if target_type in (F.dir, "root"):
                    if name not in target_file[F.children]:
                        target_file[F.children][name] = source_file
                        del source_parent[F.children][name]
                        if self.editlog:
                            self.editlog.writeline({F.cmd: C.move, F.source_path: source_path, F.target_path: target_path})
                        result = True
                    else:
                        raise SameNameExistsError("same file name exists: %s" % name)
                else:
                    raise TargetPathMustDirectoryError("target path must be directory: %s" % target_path)
            else:
                raise TargetPathNotExistsError("target path not exists: %s" % target_path)
        else:
            raise SourcePathNotExistsError("source path not exists: %s" % source_path)
        return result

    def copy(self, source_path, target_path):
        result = False
        _, name = os.path.split(source_path)
        source_exists, _, source_file, source_parent = self.get_info(source_path)
        if source_exists:
            target_exists, target_type, target_file, _ = self.get_info(target_path)
            if target_exists:
                if target_type == F.dir:
                    if name not in target_file[F.children]:
                        target_file[F.children][name] = source_file
                        if self.editlog:
                            self.editlog.writeline({F.cmd: C.copy, F.source_path: source_path, F.target_path: target_path})
                        result = True
                    else:
                        raise SameNameExistsError("same file name exists: %s" % name)
                else:
                    raise TargetPathMustDirectoryError("target path must be directory: %s" % target_path)
            else:
                raise TargetPathNotExistsError("target path not exists: %s" % target_path)
        else:
            raise SourcePathNotExistsError("source path not exists: %s" % source_path)
        return result

    def list_dir(self, directory_path, offset = 0, limit = 0, recursive = False, include_file = True, include_directory = True):
        result = {"files": []}
        if offset < 0:
            offset = 0
        if limit < 0:
            limit = 0
        result["offset"] = offset
        result["limit"] = limit
        files = []
        dirs = []
        exists, file_type, file, _ = self.get_info(directory_path)
        if exists and file_type in (F.dir, "root"):
            if recursive:
                result = file[F.children]
            else:
                names = list(file[F.children].keys())
                names.sort()
                for name in names:
                    c = file[F.children][name]
                    file_type = c[F.type]
                    child = {
                        "name": name,
                    }
                    if include_file and file_type == F.file:
                        file_id = file[F.children][name][F.id]
                        file_info = self.files[file_id]
                        child["type"] = "file"
                        child["size"] = file_info["size"]
                        child["id"] = file_id
                        if "ctime" in file_info:
                            child["ctime"] = file_info["ctime"]
                        if "mtime" in file_info:
                            child["mtime"] = file_info["mtime"]
                        if "current_replica" in file_info:
                            child["current_replica"] = file_info["current_replica"]
                        if "replica" in file_info:
                            child["replica"] = file_info["replica"]
                        files.append(child)
                    elif include_directory and file_type == F.dir:
                        child["type"] = "directory"
                        child["size"] = 0
                        child["id"] = ""
                        if "ctime" in c[F.info]:
                            child["ctime"] = c[F.info]["ctime"]
                        if "mtime" in c[F.info]:
                            child["mtime"] = c[F.info]["mtime"]
                        dirs.append(child)
                result["files"].extend(dirs)
                result["files"].extend(files)
                result["total"] = len(result["files"])
                if offset > 0 and limit > 0:
                    result["files"] = result["files"][offset:offset + limit]
                elif offset > 0:
                    result["files"] = result["files"][offset:]
                elif limit > 0:
                    result["files"] = result["files"][:limit]
            LOG.debug("list_dir: %s", result)
        return result

    def exists(self, file_path):
        result, _, _, _ = self.get_info(file_path)
        return result

    def makedir(self, directory_path, directory_info = None, recover = False):
        result = False
        path_parts = splitall(directory_path)
        if path_parts[0] != "/":
            raise InvalidValueError("must be absolute path: %s" % directory_path)
        else:
            new_directory_flag = False
            now = int(time.time())
            current_root = self.tree
            if directory_info is None:
                directory_info = {"ctime": now, "mtime": now}
            last_idx = len(path_parts) - 2
            for n, dir_name in enumerate(path_parts[1:]):
                if dir_name not in current_root[F.children]:
                    if n != last_idx:
                        raise ParentDirectoryNotExistsError("parent directory not exists: %s" % os.path.join(*path_parts[:n + 2]))
                    else:
                        current_root[F.children][dir_name] = {F.type: F.dir, F.children: {}, F.info: directory_info}
                        new_directory_flag = True
                        if self.editlog:
                            self.editlog.writeline({F.cmd: C.makedir, F.path: directory_path, F.info: directory_info})
                else:
                    child = current_root[F.children][dir_name]
                    if child[F.type] == F.file:
                        raise SameNameFileExistsError("same file name exists: %s" % dir_name)
                current_root = current_root[F.children][dir_name]
            update_root = self.tree
            if new_directory_flag and not recover:
                self.update_parent_dirs(os.path.join(*path_parts[:-1]), directory_info)
            result = current_root
        return result

    def makedirs(self, directory_path, recover = False):
        result = False
        path_parts = splitall(directory_path)
        if path_parts[0] != "/":
            raise InvalidValueError("must be absolute path: %s" % directory_path)
        else:
            current_root = self.tree
            for n, dir_name in enumerate(path_parts[1:]):
                current_root = self.makedir(os.path.join(*path_parts[:n + 2]), recover = recover)
            result = current_root
        return result

    def update_parent_dirs(self, directory_path, info = None):
        result = False
        path_parts = splitall(directory_path)
        if path_parts[0] != "/":
            raise InvalidValueError("must be absolute path: %s" % directory_path)
        else:
            if info is None:
                info = {"mtime": int(time.time())}
            else:
                info = {"mtime": info["mtime"]}
            update_root = self.tree
            for dir_name in path_parts[1:]:
                if dir_name in update_root[F.children]:
                    update_root[F.children][dir_name][F.info]["mtime"] = info["mtime"]
                update_root = update_root[F.children][dir_name]
            if self.editlog:
                self.editlog.writeline({F.cmd: C.update_parent_dirs, F.path: directory_path, F.info: info})
            result = True
        return result

    def isdir(self, directory_path):
        result = False
        exists, file_type, _, _ = self.get_info(directory_path)
        if exists and file_type == F.dir:
            result = True
        return result

    def isfile(self, file_path):
        result = False
        exists, file_type, _, _ = self.get_info(file_path)
        if exists and file_type == F.file:
            result = True
        return result

    def get_info(self, file_path):
        result = [True, "", {}, None]
        path_parts = splitall(file_path)
        if path_parts[0] != "/":
            raise InvalidValueError("must be absolute path: %s" % file_path)
        else:
            parent = None
            current_root = self.tree
            for name in path_parts[1:]:
                if name not in current_root[F.children]:
                    result[0] = False
                    break
                else:
                    parent = current_root
                    current_root = current_root[F.children][name]
            if result[0]:
                result[1] = current_root[F.type]
                result[2] = current_root
                result[3] = parent
        return result

    @gen.coroutine
    def load_fsimage(self):
        result = False
        try:
            LOG.info("loading fsimage ...")
            fsimage_path = os.path.join(CONFIG["data_path"], "fsimage")
            fsimage = AppendLogJson(fsimage_path)
            n = 0
            for line in fsimage.iterlines():
                if n >= 100:
                    n = 0
                    yield gen.moment
                if line[F.cmd] == C.create:
                    self.create(line[F.path], line[F.info], recover = True)
                elif line[F.cmd] == C.makedir:
                    self.makedir(line[F.path], line[F.info], recover = True)
                elif line[F.cmd] == C.makedirs: # just for loading old fsimage, need convert fsimage first "md" to "mds"
                    self.makedirs(line[F.path], recover = True)
                n += 1
            result = True
        except Exception as e:
            LOG.exception(e)
        raise gen.Return(result)

    @gen.coroutine
    def load_editlog(self):
        result = False
        try:
            LOG.info("loading editlog ...")
            editlog_path = os.path.join(CONFIG["data_path"], "editlog")
            editlog = AppendLogJson(editlog_path)
            n = 0
            for line in editlog.iterlines():
                if n >= 100:
                    n = 0
                    yield gen.moment
                if line[F.cmd] == C.create:
                    self.create(line[F.path], line[F.info], recover = True)
                elif line[F.cmd] == C.makedir:
                    self.makedir(line[F.path], line[F.info], recover = True)
                elif line[F.cmd] == C.makedirs:
                    self.makedirs(line[F.path], recover = True)
                elif line[F.cmd] == C.rename:
                    self.rename(line[F.path], line[F.new_name], recover = True)
                elif line[F.cmd] == C.delete:
                    self.delete(line[F.path], recover = True)
                elif line[F.cmd] == C.move:
                    self.move(line[F.source_path], line[F.target_path])
                elif line[F.cmd] == C.copy:
                    self.copy(line[F.source_path], line[F.target_path])
                elif line[F.cmd] == C.update_file_info:
                    self.update_file_info(line[F.path], line[F.info], recover = True)
                elif line[F.cmd] == C.update_parent_dirs:
                    self.update_parent_dirs(line[F.path], line[F.info])
                n += 1
            result = True
        except Exception as e:
            LOG.exception(e)
        raise gen.Return(result)

    @gen.coroutine
    def dump_fsimage(self):
        result = False
        try:
            LOG.info("dumping fsimage ...")
            new_fsimage_path = os.path.join(CONFIG["data_path"], "fsimage.new")
            fsimage_path = os.path.join(CONFIG["data_path"], "fsimage")
            old_fsimage_path = os.path.join(CONFIG["data_path"], "fsimage.old")
            editlog_path = os.path.join(CONFIG["data_path"], "editlog")
            LOG.debug("new fsimage: %s", new_fsimage_path)
            self.new_fsimage = AppendLogJson(new_fsimage_path)
            yield self.dump_files("/", self.tree)
            if os.path.exists(old_fsimage_path):
                os.remove(old_fsimage_path)
            if os.path.exists(fsimage_path):
                os.rename(fsimage_path, old_fsimage_path)
            if os.path.exists(new_fsimage_path):
                os.rename(new_fsimage_path, fsimage_path)
            if os.path.exists(editlog_path):
                os.remove(editlog_path)
            result = True
        except Exception as e:
            LOG.exception(e)
        raise gen.Return(result)

    @gen.coroutine
    def dump_files(self, file_path, file):
        if F.children in file and file[F.children]:
            if file_path != "/":
                LOG.debug("find directory[%s]: %s", file_path, file)
                self.new_fsimage.writeline({F.cmd: C.makedir, F.path: file_path, F.info: file[F.info]})
            for name in file[F.children]:
                child = file[F.children][name]
                yield self.dump_files(os.path.join(file_path, name), child)
        else:
            if file[F.type] == F.file:
                file_id = file[F.id]
                file_info = self.files[file_id]
                LOG.debug("find file[%s]: %s", file_path, file)
                self.new_fsimage.writeline({F.cmd: C.create, F.path: file_path, F.info: file_info})
            elif file[F.type] == F.dir:
                LOG.debug("find directory[%s]: %s", file_path, file)
                self.new_fsimage.writeline({F.cmd: C.makedir, F.path: file_path, F.info: file[F.info]})

    def close(self):
        pass



# -*- coding: utf-8 -*-

import logging
from uuid import uuid4

import tornado.tcpserver
from tornado import gen
from tornado.ioloop import IOLoop
from tornado_discovery.connection import BaseConnection
from tornado_discovery.listener import BaseListener
from tornado_discovery.common import crc32sum, Command, Status, Message

from litedfs.name.models.data_nodes import DataNodes
from litedfs.name.utils.common import OperationError

LOG = logging.getLogger(__name__)


class Connection(BaseConnection):
    clients_dict = {}
    id_compress = {}
    id_decompress = {}
    tasks = {}

    def __init__(self, stream, address):
        super(Connection, self).__init__(stream, address)

    @classmethod
    def push_task(cls, id, task):
        if id in cls.tasks:
            cls.tasks[id].append(task)
        else:
            cls.tasks[id] = [task]
        LOG.debug("push task, id: %s, task: %s", id, task)

    @classmethod
    def load_node_ids(cls):
        nodes_info = DataNodes.instance().list()
        for node in nodes_info["data_nodes"]:
            cls.id_compress[node["node_id"]] = node["id"]
            cls.id_decompress[node["id"]] = node["node_id"]
        LOG.debug("id_compress: %s, id_decompress: %s", cls.id_compress, cls.id_decompress)

    @classmethod
    def get_node_infos(cls, current = None, without_full_node = False):
        result = {}
        for node_id in cls.clients_dict:
            node = cls.clients_dict[node_id]
            if node is not current:
                if without_full_node:
                    if not node.info["storage_full"]:
                        result[node.id] = [node.info["http_host"], node.info["http_port"], node.info["storage_full"]]
                else:
                    result[node.id] = [node.info["http_host"], node.info["http_port"], node.info["storage_full"]]
        return result

    @gen.coroutine
    def _on_connect(self):
        try:
            while True:
                refuse_connect_flag = False
                data = yield self.read_message()
                send_data = {}
                # register
                if "command" in data and data["command"] == Command.register:
                    self.info = data["data"]
                    if self.info["http_host"] == "0.0.0.0":
                        self.info["http_host"] = self._address[0]
                    # no node_id
                    if "node_id" not in self.info or self.info["node_id"] == None:
                        node_id = str(uuid4())
                        send_data = {
                            "command": Command.register,
                            "data": {
                                "status": Status.success,
                                "message": Status.success,
                                "node_id": node_id,
                            }
                        }
                        self.info["node_id"] = node_id
                        success = DataNodes.instance().add(node_id, info = self.info)
                        if not success:
                            send_data = {
                                "command": Command.register,
                                "data": {
                                    "status": Status.failure,
                                    "message": "add data node into data_nodes database failed",
                                }
                            }
                        else:
                            node = DataNodes.instance().get(node_id)
                            self.id = node["id"]
                            Connection.id_decompress[self.id] = self.info["node_id"]
                            Connection.id_compress[self.info["node_id"]] = self.id
                            if self.info["node_id"] not in Connection.clients_dict:
                                Connection.clients_dict[self.info["node_id"]] = self
                            self._status = Status.registered
                    # register with node_id
                    else:
                        send_data = {
                            "command": Command.register,
                            "data": {
                                "status": Status.success,
                                "message": Status.success,
                                "node_id": self.info["node_id"],
                            }
                        }
                        node = DataNodes.instance().get(self.info["node_id"])
                        if node is False:
                            send_data = {
                                "command": Command.register,
                                "data": {
                                    "status": Status.failure,
                                    "message": "data node retrive from data_nodes database failed",
                                }
                            }
                        elif node is None:
                            send_data = {
                                "command": Command.register,
                                "data": {
                                    "status": Status.failure,
                                    "message": "data node not exists in data_nodes database",
                                }
                            }
                        else:
                            self.id = node["id"]
                            Connection.id_decompress[self.id] = self.info["node_id"]
                            Connection.id_compress[self.info["node_id"]] = self.id
                            DataNodes.instance().update(self.info["node_id"], {"info": self.info})
                            if self.info["node_id"] not in Connection.clients_dict:
                                Connection.clients_dict[self.info["node_id"]] = self
                            self._status = Status.registered
                elif "command" in data and data["command"] == Command.heartbeat:
                    self.info = data["data"]
                    if self.info["http_host"] == "0.0.0.0":
                        self.info["http_host"] = self._address[0]
                    if self._status == Status.registered:
                        send_data = {
                            "command": Command.heartbeat,
                            "data": {
                                "data_nodes": Connection.get_node_infos(current = self),
                                "status": Status.success,
                                "message": Status.success,
                            }
                        }
                        if self.id in Connection.tasks and Connection.tasks[self.id]:
                            if not self.info["task_queue_full"]:
                                task = Connection.tasks[self.id].pop(0)
                                send_data["data"]["task"] = task
                        if self._heartbeat_timeout:
                            IOLoop.instance().remove_timeout(self._heartbeat_timeout)
                        self._heartbeat_timeout = IOLoop.instance().add_timeout(
                            IOLoop.time(IOLoop.instance()) + self.info["heartbeat_timeout"],
                            self._remove_connection
                        )
                    else:
                        send_data = {
                            "command": Command.heartbeat,
                            "data": {
                                "status": Status.failure,
                                "message": "invalid node_id: %s" % self.info["node_id"],
                            }
                        }
                        refuse_connect_flag = True
                else:
                    LOG.error("Client(%s) invaild message error: %s", self._address, data)
                    send_data = {
                        "command": Command.error,
                        "data": {
                            "status": Status.failure,
                            "message": "Unknown Command!",
                        }
                    }
                self.send_message(send_data, refuse_connect_flag = refuse_connect_flag)
        except tornado.iostream.StreamClosedError:
            LOG.info("Closed: %s", self._address)
        except Exception as e:
            LOG.exception(e)

    def _remove_connection(self):
        if self in BaseConnection.clients:
            if "node_id" in self.info and self.info["node_id"] in Connection.clients_dict:
                del Connection.clients_dict[self.info["node_id"]]
            BaseConnection.clients.remove(self)
        self._stream.close()
        LOG.warning("Client(%s) node_id: %s heartbeat_timeout", self._address, self.info["node_id"])

    def _refuse_connect(self):
        if self._heartbeat_timeout:
            IOLoop.instance().remove_timeout(self._heartbeat_timeout)
        if self in BaseConnection.clients:
            if "node_id" in self.info and self.info["node_id"] in Connection.clients_dict:
                del Connection.clients_dict[self.info["node_id"]]
            BaseConnection.clients.remove(self)
        self._stream.close()
        LOG.warning("Refuse(%s) node_id: %s connect", self._address, self.info["node_id"])

    def _on_close(self):
        if self._heartbeat_timeout:
            IOLoop.instance().remove_timeout(self._heartbeat_timeout)
        if self in BaseConnection.clients:
            if "node_id" in self.info and self.info["node_id"] in Connection.clients_dict:
                del Connection.clients_dict[self.info["node_id"]]
            BaseConnection.clients.remove(self)
        self._stream.close()
        LOG.info("Client(%s) closed", self._address)


class DiscoveryListener(BaseListener):
    def __init__(self, connection_cls, ssl_options = None, **kwargs):
        LOG.info("DiscoveryListener start")
        self.connection_cls = connection_cls
        self.connection_cls.load_node_ids()
        tornado.tcpserver.TCPServer.__init__(self, ssl_options = ssl_options, **kwargs)

    def handle_stream(self, stream, address):
        LOG.debug("Incoming connection from %r", address)
        self.connection_cls(stream, address)



# -*- coding: utf-8 -*-

import os
import json
import time
import shutil
import logging
from pathlib import Path

from tornado import web
from tornado import gen

from litedfs.tool.viewer.handlers.base import BaseHandler, BaseSocketHandler
from litedfs.tool.viewer.utils.common import listdir, joinpath, splitpath, list_storage, Command
from litedfs.tool.viewer.utils.task_cache import TaskCache
from litedfs.tool.viewer.utils.remote_storage import RemoteStorage
from litedfs.tool.viewer.config import CONFIG

LOG = logging.getLogger("__name__")


class StorageHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        self.render(
            "storage/storage.html",
            current_nav = "storage",
            manager_host = "%s:%s" % (
                self.get_name_http_host(),
                CONFIG["name_http_port"])
        )


def send_msg(msg, handler):
    try:
        handler.write_message(msg)
    except Exception as e:
        LOG.exception(e)


def send_msgs(msg, handlers):
    try:
        for handler in handlers:
            handler.write_message(msg)
    except Exception as e:
        LOG.exception(e)


class LocalSocketHandler(BaseSocketHandler):
    socket_handlers = set()

    @classmethod
    def send_msgs(cls, msg):
        try:
            for handler in cls.socket_handlers:
                handler.write_message(msg)
        except Exception as e:
            LOG.exception(e)

    def open(self):
        self.home_path = str(Path.home())
        self.client = RemoteStorage(CONFIG["name_http_host"], CONFIG["name_http_port"])
        self.clipboard = {}
        if self not in LocalSocketHandler.socket_handlers:
            LocalSocketHandler.socket_handlers.add(self)
            LOG.info("storage websocket len: %s", len(LocalSocketHandler.socket_handlers))
        else:
            LOG.info("storage websocket len: %s", len(LocalSocketHandler.socket_handlers))
        data = list_storage(self.home_path, self.home_path, sort_by = "name", desc = False, offset = 0, limit = 100)
        data["cmd"] = "init"
        send_msg(json.dumps(data), self)

    def on_close(self):
        LocalSocketHandler.socket_handlers.remove(self)
        LOG.info("storage websocket len: %s", len(LocalSocketHandler.socket_handlers))

    def on_message(self, msg):
        msg = json.loads(msg)
        LOG.debug("msg: %s", msg)
        data = {}
        try:
            if msg["cmd"] == Command.cd:
                cd_path = joinpath(msg["dir_path"])
                data = list_storage(self.home_path, cd_path, sort_by = "name", desc = False, offset = msg["offset"], limit = msg["limit"])
                data["cmd"] = "init"
                send_msg(json.dumps(data), self)
            elif msg["cmd"] == Command.refresh:
                dir_path = joinpath(msg["dir_path"])
                data = list_storage(self.home_path, dir_path, sort_by = "name", desc = False, offset = msg["offset"], limit = msg["limit"])
                data["cmd"] = "init"
                send_msg(json.dumps(data), self)
            elif msg["cmd"] == Command.change_page:
                dir_path = joinpath(msg["dir_path"])
                data = list_storage(self.home_path, dir_path, sort_by = "name", desc = False, offset = msg["offset"], limit = msg["limit"])
                data["cmd"] = "init"
                send_msg(json.dumps(data), self)
            elif msg["cmd"] == Command.rename:
                dir_path = joinpath(msg["dir_path"])
                old_name = msg["old_name"]
                new_name = msg["new_name"]
                if new_name != "" and new_name != old_name:
                    old_path = os.path.join(dir_path, old_name)
                    new_path = os.path.join(dir_path, new_name)
                    if os.path.exists(new_path):
                        data["cmd"] = "warning"
                        data["info"] = "File [%s] already exists!" % new_path
                    else:
                        os.rename(old_path, new_path)
                        data = list_storage(self.home_path, dir_path, sort_by = "name", desc = False, offset = msg["offset"], limit = msg["limit"])
                        data["cmd"] = "init"
                    send_msg(json.dumps(data), self)
            elif msg["cmd"] == Command.mkdir:
                dir_path = joinpath(msg["dir_path"])
                dir_name = msg["name"]
                if dir_name != "":
                    new_path = os.path.join(dir_path, dir_name)
                    if os.path.exists(new_path):
                        data["cmd"] = "warning"
                        data["info"] = "Directory [%s] already exists!" % new_path
                    else:
                        os.mkdir(new_path)
                        data = list_storage(self.home_path, dir_path, sort_by = "name", desc = False, offset = msg["offset"], limit = msg["limit"])
                        data["cmd"] = "init"
                    send_msg(json.dumps(data), self)
            elif msg["cmd"] == Command.delete:
                msg["socket_handler"] = self
                TaskCache.push(msg)
            elif msg["cmd"] == Command.copy:
                dir_path = joinpath(msg["dir_path"])
                files = msg["files"]
                dirs = msg["dirs"]
                self.clipboard["type"] = Command.copy
                self.clipboard["dir_path"] = dir_path
                self.clipboard["files"] = files
                self.clipboard["dirs"] = dirs
                data["cmd"] = "paste"
                send_msg(json.dumps(data), self)
            elif msg["cmd"] == Command.cut:
                dir_path = joinpath(msg["dir_path"])
                files = msg["files"]
                dirs = msg["dirs"]
                self.clipboard["type"] = Command.cut
                self.clipboard["dir_path"] = dir_path
                self.clipboard["files"] = files
                self.clipboard["dirs"] = dirs
                data["cmd"] = "paste"
                send_msg(json.dumps(data), self)
            elif msg["cmd"] == Command.paste:
                msg["socket_handler"] = self
                msg["clipboard"] = self.clipboard
                TaskCache.push(msg)
            elif msg["cmd"] == Command.upload:
                msg["socket_handler"] = self
                TaskCache.push(msg)
        except Exception as e:
            LOG.exception(e)
            data["cmd"] = "error"
            data["info"] = str(e)
            send_msg(json.dumps(data), self)


class RemoteSocketHandler(BaseSocketHandler):
    socket_handlers = set()

    @classmethod
    def send_msgs(cls, msg):
        try:
            for handler in cls.socket_handlers:
                handler.write_message(msg)
        except Exception as e:
            LOG.exception(e)

    def open(self):
        self.home_path = "/"
        self.client = RemoteStorage(CONFIG["name_http_host"], CONFIG["name_http_port"])
        self.clipboard = {}
        if self not in RemoteSocketHandler.socket_handlers:
            RemoteSocketHandler.socket_handlers.add(self)
            LOG.info("remote storage websocket len: %s", len(RemoteSocketHandler.socket_handlers))
        else:
            LOG.info("remote storage websocket len: %s", len(RemoteSocketHandler.socket_handlers))
        data = self.client.list_storage(self.home_path, self.home_path, sort_by = "name", desc = False, offset = 0, limit = 100)
        if data:
            data["cmd"] = "init"
        else:
            data["cmd"] = "error"
            data["info"] = "Remote storage is offline"
        send_msg(json.dumps(data), self)

    def on_close(self):
        RemoteSocketHandler.socket_handlers.remove(self)
        LOG.info("remote storage websocket len: %s", len(RemoteSocketHandler.socket_handlers))

    def on_message(self, msg):
        msg = json.loads(msg)
        LOG.debug("msg: %s", msg)
        data = {}
        try:
            if msg["cmd"] == Command.cd:
                cd_path = joinpath(msg["dir_path"])
                data = self.client.list_storage(self.home_path, cd_path, sort_by = "name", desc = False, offset = msg["offset"], limit = msg["limit"])
                if data:
                    data["cmd"] = "init"
                else:
                    data["cmd"] = "error"
                    data["info"] = "Remote storage is offline"
                send_msg(json.dumps(data), self)
            elif msg["cmd"] == Command.refresh:
                dir_path = joinpath(msg["dir_path"])
                data = self.client.list_storage(self.home_path, dir_path, sort_by = "name", desc = False, offset = msg["offset"], limit = msg["limit"])
                if data:
                    data["cmd"] = "init"
                else:
                    data["cmd"] = "error"
                    data["info"] = "Remote storage is offline"
                send_msg(json.dumps(data), self)
            elif msg["cmd"] == Command.change_page:
                dir_path = joinpath(msg["dir_path"])
                data = self.client.list_storage(self.home_path, dir_path, sort_by = "name", desc = False, offset = msg["offset"], limit = msg["limit"])
                if data:
                    data["cmd"] = "init"
                else:
                    data["cmd"] = "error"
                    data["info"] = "Remote storage is offline"
                send_msg(json.dumps(data), self)
            elif msg["cmd"] == Command.rename:
                dir_path = joinpath(msg["dir_path"])
                old_name = msg["old_name"]
                new_name = msg["new_name"]
                if new_name != "" and new_name != old_name:
                    old_path = os.path.join(dir_path, old_name)
                    self.client.rename(old_path, new_name)
                    data = self.client.list_storage(self.home_path, dir_path, sort_by = "name", desc = False, offset = msg["offset"], limit = msg["limit"])
                    if data:
                        data["cmd"] = "init"
                    else:
                        data["cmd"] = "error"
                        data["info"] = "Remote storage is offline"
                    send_msg(json.dumps(data), self)
            elif msg["cmd"] == Command.mkdir:
                dir_path = joinpath(msg["dir_path"])
                dir_name = msg["name"]
                if dir_name != "":
                    new_path = os.path.join(dir_path, dir_name)
                    self.client.mkdir(new_path)
                    data = self.client.list_storage(self.home_path, dir_path, sort_by = "name", desc = False, offset = msg["offset"], limit = msg["limit"])
                    if data:
                        data["cmd"] = "init"
                    else:
                        data["cmd"] = "error"
                        data["info"] = "Remote storage is offline"
                    send_msg(json.dumps(data), self)
            elif msg["cmd"] == Command.delete:
                msg["socket_handler"] = self
                msg["cmd"] = Command.remote_delete
                TaskCache.push(msg)
            elif msg["cmd"] == Command.cut:
                dir_path = joinpath(msg["dir_path"])
                files = msg["files"]
                dirs = msg["dirs"]
                self.clipboard["type"] = Command.cut
                self.clipboard["dir_path"] = dir_path
                self.clipboard["files"] = files
                self.clipboard["dirs"] = dirs
                data["cmd"] = "paste"
                send_msg(json.dumps(data), self)
            elif msg["cmd"] == Command.paste:
                msg["socket_handler"] = self
                msg["clipboard"] = self.clipboard
                msg["cmd"] = Command.remote_paste
                TaskCache.push(msg)
            elif msg["cmd"] == Command.download:
                msg["socket_handler"] = self
                TaskCache.push(msg)
            elif msg["cmd"] == Command.update:
                msg["socket_handler"] = self
                TaskCache.push(msg)
            elif msg["cmd"] == Command.preview:
                msg["socket_handler"] = self
                TaskCache.push(msg)
        except Exception as e:
            LOG.exception(e)
            data["cmd"] = "error"
            data["info"] = str(e)
            send_msg(json.dumps(data), self)



# -*- coding: utf-8 -*-

import os
import time
import json
import hashlib
import logging
import datetime
import uuid
import mimetypes
from functools import partial

from tornado import ioloop
from tornado import gen
from tornado.web import HTTPError
import psutil

from litedfs.data.config import CONFIG

LOG = logging.getLogger(__name__)

BUF_SIZE = 65536


class Servers(object):
    HTTP_SERVER = None
    SERVERS = []
    TORNADO_INSTANCE = None


async def shutdown():
    LOG.info("Stopping Service(%s:%s)", CONFIG["http_host"], CONFIG["http_port"])
    if Servers.HTTP_SERVER:
        Servers.HTTP_SERVER.stop()
        LOG.info("Stop http server!")
    for s in Servers.SERVERS:
        if hasattr(s, "close"):
            s.close()
        elif hasattr(s, "stop"):
            s.stop()
        if hasattr(s, "name"):
            LOG.info("Stop %s server!", s.name)
        else:
            LOG.info("Stop nameless server!")
    await gen.sleep(1)
    LOG.info("Will shutdown ...")
    ioloop.IOLoop.current().stop()


def sig_handler(sig, frame):
    LOG.warning("sig_handler Caught signal: %s", sig)
    ioloop.IOLoop.current().add_callback_from_signal(shutdown)


class Errors(object):
    OK = "ok"
    errors = {
        "ServerException": {"name": "ServerException", "message": "server exception"},
        "InvalidParameters": {"name": "InvalidParameters", "message": "invalid parameters"},
        "OperationFailed": {"name": "OperationFailed", "message": "operation failed"},
        "OperationRunning": {"name": "OperationRunning", "message": "operation running"},
        "BlockNotExists": {"name": "BlockNotExists", "message": "block not exists"},
        "ChecksumFailed": {"name": "ChecksumFailed", "message": "checksum failed"},
        "ReplicateBlockFailed": {"name": "ReplicateBlockFailed", "message": "replicate block failed"},
    }

    @classmethod
    def set_result_error(cls, error_name, result, message = ""):
        if error_name in cls.errors:
            result["result"] = error_name
            if message == "":
                result["message"] = cls.errors[error_name]["message"]
            else:
                result["message"] = message
        else:
            result["result"] = "UnknownError"
            result["message"] = "unknown error"


class Stage(object):
    pending = "pending"
    running = "running"
    finished = "finished"


class Status(object):
    fail = "fail"
    success = "success"
    kill = "kill"
    cancel = "cancel"
    terminate = "terminate"


class Signal(object):
    kill = -9
    terminate = -15
    stop = -15
    cancel = -50


class JSONLoadError(Exception):
    def __init__(self, message):
        self.message = message


class MetaNotDictError(Exception):
    def __init__(self, message):
        self.message = message


def file_sha1sum(file_path):
    sha1 = hashlib.sha1()
    with open(file_path, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha1.update(data)
    return sha1.hexdigest()


def file_md5sum(file_path):
    md5 = hashlib.md5()
    with open(file_path, 'rb') as fp:
        while True:
            data = fp.read(BUF_SIZE)
            if not data:
                break
            md5.update(data)
    return md5.hexdigest()


def bytes_md5sum(b):
    md5 = hashlib.md5()
    md5.update(b)
    return md5.hexdigest()


def splitall(path):
    allparts = []
    while True:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            if parts[1]:
                allparts.insert(0, parts[1])
    return allparts


def get_workspace_path(create_at, task_id = None, action_name = None):
    result = ""
    date_create_at = datetime.datetime.strptime(create_at, "%Y-%m-%d %H:%M:%S.%f")
    date_directory_name = date_create_at.strftime("%Y-%m-%d")
    if task_id and action_name:
        result = os.path.join(CONFIG["data_path"], "tmp", "workspace", date_directory_name, task_id[:2], task_id[2:4], task_id, action_name)
    elif task_id:
        result = os.path.join(CONFIG["data_path"], "tmp", "workspace", date_directory_name, task_id[:2], task_id[2:4], task_id)
    else:
        result = os.path.join(CONFIG["data_path"], "tmp", "workspace", date_directory_name)
    return result


def init_storage():
    directories = [
        os.path.join(CONFIG["data_path"], "files"),
        os.path.join(CONFIG["data_path"], "tmp", "download"),
    ]
    for d in directories:
        if not os.path.exists(d) or not os.path.isdir(d):
            os.makedirs(d)


def delete_file(name):
    try:
        dir_path = os.path.join(CONFIG["data_path"], "files", name[:2], name[2:4])
        if os.path.exists(dir_path):
            files = os.listdir(dir_path)
            for file in files:
                if file.startswith(name):
                    os.remove(os.path.join(dir_path, file))
            files = os.listdir(dir_path)
            if len(files) == 0:
                os.rmdir(dir_path)
                dir_parent_path = os.path.split(dir_path)[0]
                files = os.listdir(dir_parent_path)
                if len(files) == 0:
                    os.rmdir(dir_parent_path)
    except Exception as e:
        LOG.exception(e)


def delete_block(name, block):
    try:
        dir_path = os.path.join(CONFIG["data_path"], "files", name[:2], name[2:4])
        file_path = os.path.join(dir_path, "%s_%s.blk" % (name, block))
        tmp_file_path = os.path.join(dir_path, "%s_%s.blk.tmp" % (name, block))
        check_file_path = os.path.join(dir_path, "%s_%s.chk" % (name, block))
        if os.path.exists(file_path):
            os.remove(file_path)
        if os.path.exists(tmp_file_path):
            os.remove(tmp_file_path)
        if os.path.exists(check_file_path):
            os.remove(check_file_path)
        files = os.listdir(dir_path)
        if len(files) == 0:
            os.rmdir(dir_path)
            dir_parent_path = os.path.split(dir_path)[0]
            files = os.listdir(dir_parent_path)
            if len(files) == 0:
                os.rmdir(dir_parent_path)
    except Exception as e:
        LOG.exception(e)


def body_producer(boundary, files, params, write):
    if not isinstance(files, dict):
        raise HTTPError("files must be dict")
    if not isinstance(params, dict):
        raise HTTPError("params must be dict")

    boundary_bytes = boundary.encode()

    for file_name in files:
        file = files[file_name]
        file_name_bytes = file_name.encode()
        write(b'--%s\r\n' % (boundary_bytes, ))
        write(b'Content-Disposition: form-data; name="%s"; filename="%s"\r\n' %
              (file_name.encode(), file_name_bytes))

        mtype = mimetypes.guess_type(file_name)[0] or 'application/octet-stream'
        write(b'Content-Type: %s\r\n' % (mtype.encode(), ))
        write(b'\r\n')
        while True:
            # 64k at a time.
            chunk = file.read(64 * 1024)
            if not chunk:
                break
            write(chunk)

        write(b'\r\n')

    for arg_name in params:
        value = params[arg_name]
        write(b'--%s\r\n' % (boundary_bytes, ))
        write(b'Content-Disposition: form-data; name="%s"\r\n\r\n%s\r\n' %
              (arg_name.encode(), value.encode()))

    write(b'--%s--\r\n' % (boundary_bytes, ))


def async_post(async_client, url, files, params):
    boundary = uuid.uuid4().hex
    headers = {'Content-Type': 'multipart/form-data; boundary=%s' % boundary}
    producer = partial(body_producer, boundary, files, params)
    response = async_client.fetch(
        url,
        method = 'POST',
        headers = headers,
        body_producer = producer
    )
    return response


def size_pretty(size):
    result = ""
    try:
        if size > 1024*1014*1024*1024:
            result = "%.3f T"%(size/1024.0/1024.0/1024.0/1024.0)
        elif size > 1024*1014*1024:
            result = "%.3f G"%(size/1024.0/1024.0/1024.0)
        elif size > 1024*1024:
            result = "%.3f M"%(size/1024.0/1024.0)
        elif size > 1024:
            result = "%.3f K"%(size/1024.0)
        else:
            result = "%d B"%size
    except Exception as e:
        LOG.exception(e)
        result = "0 B"
    return result


def disk_usage(dir_path = None):
    if not dir_path:
        dir_path = CONFIG["data_path"]
    usage = psutil.disk_usage(dir_path)
    return {
        "total": usage.total,
        "used": usage.used,
        "free": usage.free,
        "percent": usage.percent
    }



# -*- coding: utf-8 -*-

import json
import time
import logging

from tornado import web
from tornado import gen

from litedfs.name.handlers.base import BaseHandler, BaseSocketHandler
from litedfs.name.models.data_nodes import DataNodes
from litedfs.name.utils.listener import Connection
from litedfs.name.utils.common import Errors, list_sort
from litedfs.version import __version__
from litedfs.name.config import CONFIG

LOG = logging.getLogger("__name__")


class AboutHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"message": "LiteDFS name service"}
        self.write(result)
        self.finish()


class ClusterInfoHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK, "version": __version__}
        try:
            info = {
                "number_of_nodes": 0,
                "number_of_online_nodes": 0,
                "number_of_offline_nodes": 0,
                "online_nodes": [],
                "offline_nodes": [],
            }
            for node in Connection.clients:
                node.info["id"] = Connection.id_compress[node.info["node_id"]]
                info["online_nodes"].append(node.info)
                info["number_of_nodes"] += 1
                info["number_of_online_nodes"] += 1
            info["online_nodes"] = list_sort(info["online_nodes"], "id")
            nodes_info = DataNodes.instance().list()
            for node in nodes_info["data_nodes"]:
                if node["node_id"] not in Connection.clients_dict:
                    node["info"]["id"] = node["id"]
                    info["offline_nodes"].append(node["info"])
                    info["number_of_offline_nodes"] += 1
            info["offline_nodes"] = list_sort(info["offline_nodes"], "id")
            result["info"] = info
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(json.dumps(result, sort_keys = True))
        self.finish()
